# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.errors import MouseTrap
import pyomo.common.unittest as unittest
from pyomo.contrib.cp.transform.logical_to_disjunctive_program import (
    LogicalToDisjunctive,
)
from pyomo.contrib.cp.transform.logical_to_disjunctive_walker import (
    LogicalToDisjunctiveVisitor,
)
from pyomo.core.expr.compare import assertExpressionsEqual
from pyomo.core.plugins.transform.logical_to_linear import (
    update_boolean_vars_from_binary,
)
from pyomo.gdp import Disjunct
from pyomo.environ import (
    atmost,
    atleast,
    exactly,
    Block,
    BooleanVar,
    Binary,
    ConcreteModel,
    Constraint,
    Expression,
    Integers,
    land,
    lnot,
    lor,
    LogicalConstraint,
    Objective,
    Param,
    SolverFactory,
    value,
    Var,
    TransformationFactory,
)

gurobi_available = (
    SolverFactory('gurobi').available(exception_flag=False)
    and SolverFactory('gurobi').license_is_valid()
)


class TestLogicalToDisjunctiveVisitor(unittest.TestCase):
    def make_model(self):
        m = ConcreteModel()

        m.a = BooleanVar()
        m.b = BooleanVar()
        m.c = BooleanVar()

        return m

    def test_logical_or(self):
        m = self.make_model()
        e = lor(m.a, m.b, m.c)

        visitor = LogicalToDisjunctiveVisitor()
        m.cons = visitor.constraints
        m.z = visitor.z_vars

        visitor.walk_expression(e)

        self.assertIs(m.a.get_associated_binary(), m.z[1])
        self.assertIs(m.b.get_associated_binary(), m.z[2])
        self.assertIs(m.c.get_associated_binary(), m.z[3])

        self.assertEqual(len(m.cons), 5)
        self.assertEqual(len(m.z), 4)
        # !z4 v a v b v c
        assertExpressionsEqual(
            self, m.cons[1].expr, 1 - m.z[4] + m.z[1] + m.z[2] + m.z[3] >= 1
        )
        # z4 v !a
        assertExpressionsEqual(self, m.cons[2].expr, m.z[4] + (1 - m.z[1]) >= 1)
        # z4 v !b
        assertExpressionsEqual(self, m.cons[3].expr, m.z[4] + (1 - m.z[2]) >= 1)
        # z4 v !c
        assertExpressionsEqual(self, m.cons[4].expr, m.z[4] + (1 - m.z[3]) >= 1)

        # z4 is constrained to be 'True'
        assertExpressionsEqual(self, m.cons[5].expr, m.z[4] >= 1)

    def test_logical_and(self):
        m = self.make_model()
        e = land(m.a, m.b, m.c)

        visitor = LogicalToDisjunctiveVisitor()
        m.cons = visitor.constraints
        m.z = visitor.z_vars

        visitor.walk_expression(e)

        self.assertIs(m.a.get_associated_binary(), m.z[1])
        self.assertIs(m.b.get_associated_binary(), m.z[2])
        self.assertIs(m.c.get_associated_binary(), m.z[3])

        self.assertEqual(len(m.cons), 5)
        self.assertEqual(len(m.z), 4)
        assertExpressionsEqual(self, m.cons[1].expr, m.z[4] <= m.z[1])
        assertExpressionsEqual(self, m.cons[2].expr, m.z[4] <= m.z[2])
        assertExpressionsEqual(self, m.cons[3].expr, m.z[4] <= m.z[3])
        assertExpressionsEqual(
            self, m.cons[4].expr, 1 - m.z[4] <= 3 - (m.z[1] + m.z[2] + m.z[3])
        )
        assertExpressionsEqual(self, m.cons[5].expr, m.z[4] >= 1)

    def test_logical_not(self):
        m = self.make_model()
        e = lnot(m.a)

        visitor = LogicalToDisjunctiveVisitor()
        m.cons = visitor.constraints
        m.z = visitor.z_vars

        visitor.walk_expression(e)
        self.assertIs(m.a.get_associated_binary(), m.z[1])
        self.assertEqual(len(m.cons), 2)
        self.assertEqual(len(m.z), 2)
        assertExpressionsEqual(self, m.cons[1].expr, m.z[2] == 1 - m.z[1])
        assertExpressionsEqual(self, m.cons[2].expr, m.z[2] >= 1)

    def test_implication(self):
        m = self.make_model()
        e = m.a.implies(m.b.land(m.c))

        visitor = LogicalToDisjunctiveVisitor()
        m.cons = visitor.constraints
        m.z = visitor.z_vars

        visitor.walk_expression(e)
        self.assertIs(m.a.get_associated_binary(), m.z[1])
        self.assertIs(m.b.get_associated_binary(), m.z[2])
        self.assertIs(m.c.get_associated_binary(), m.z[3])

        self.assertEqual(len(m.cons), 7)
        # z4 = b ^ c
        assertExpressionsEqual(self, m.cons[1].expr, m.z[4] <= m.z[2])
        assertExpressionsEqual(self, m.cons[2].expr, m.z[4] <= m.z[3])
        assertExpressionsEqual(
            self, m.cons[3].expr, 1 - m.z[4] <= 2 - (m.z[2] + m.z[3])
        )
        # z5 = a -> z4
        # which means z5 = !a v z4
        assertExpressionsEqual(
            self, m.cons[4].expr, (1 - m.z[5]) + (1 - m.z[1]) + m.z[4] >= 1
        )
        # z5 >= 1 - z1
        assertExpressionsEqual(self, m.cons[5].expr, m.z[5] + (1 - (1 - m.z[1])) >= 1)
        # z5 >= z4
        assertExpressionsEqual(self, m.cons[6].expr, m.z[5] + (1 - m.z[4]) >= 1)

        # z5 is constrained to be 'True'
        assertExpressionsEqual(self, m.cons[7].expr, m.z[5] >= 1)

    def test_equivalence(self):
        m = self.make_model()
        e = m.a.equivalent_to(m.c)

        visitor = LogicalToDisjunctiveVisitor()
        m.cons = visitor.constraints
        m.z = visitor.z_vars

        visitor.walk_expression(e)
        self.assertIs(m.a.get_associated_binary(), m.z[1])
        self.assertIs(m.c.get_associated_binary(), m.z[2])
        self.assertEqual(len(m.z), 5)
        self.assertEqual(len(m.cons), 10)

        # z[3] == !a v c
        assertExpressionsEqual(
            self, m.cons[1].expr, (1 - m.z[3]) + (1 - m.z[1]) + m.z[2] >= 1
        )
        assertExpressionsEqual(self, m.cons[2].expr, 1 - (1 - m.z[1]) + m.z[3] >= 1)
        assertExpressionsEqual(self, m.cons[3].expr, m.z[3] + (1 - m.z[2]) >= 1)

        # z[4] == a v ! c
        assertExpressionsEqual(
            self, m.cons[4].expr, (1 - m.z[4]) + (1 - m.z[2]) + m.z[1] >= 1
        )
        assertExpressionsEqual(self, m.cons[5].expr, m.z[4] + (1 - m.z[1]) >= 1)
        assertExpressionsEqual(self, m.cons[6].expr, 1 - (1 - m.z[2]) + m.z[4] >= 1)

        # z[5] == z[3] ^ z[4]
        assertExpressionsEqual(self, m.cons[7].expr, m.z[5] <= m.z[3])
        assertExpressionsEqual(self, m.cons[8].expr, m.z[5] <= m.z[4])
        assertExpressionsEqual(
            self, m.cons[9].expr, 1 - m.z[5] <= 2 - (m.z[3] + m.z[4])
        )

        assertExpressionsEqual(self, m.cons[10].expr, m.z[5] >= 1)

    def test_equivalent_to_True(self):
        m = self.make_model()
        e = m.a.equivalent_to(True)

        visitor = LogicalToDisjunctiveVisitor()
        m.cons = visitor.constraints
        m.z = visitor.z_vars

        visitor.walk_expression(e)

        self.assertIs(m.a.get_associated_binary(), m.z[1])
        self.assertEqual(len(m.z), 4)
        self.assertEqual(len(m.cons), 10)

        # z[2] == !a v True
        assertExpressionsEqual(
            self, m.cons[1].expr, (1 - m.z[2]) + (1 - m.z[1]) + 1 >= 1
        )
        assertExpressionsEqual(self, m.cons[2].expr, 1 - (1 - m.z[1]) + m.z[2] >= 1)
        assertExpressionsEqual(self, m.cons[3].expr, m.z[2] + (1 - 1) >= 1)

        # z[3] == a v ! c
        assertExpressionsEqual(self, m.cons[4].expr, (1 - m.z[3]) + m.z[1] >= 1)
        assertExpressionsEqual(self, m.cons[5].expr, m.z[3] + (1 - m.z[1]) >= 1)
        assertExpressionsEqual(self, m.cons[6].expr, m.z[3] + 1 >= 1)

        # z[4] == z[2] ^ z[3]
        assertExpressionsEqual(self, m.cons[7].expr, m.z[4] <= m.z[2])
        assertExpressionsEqual(self, m.cons[8].expr, m.z[4] <= m.z[3])
        assertExpressionsEqual(
            self, m.cons[9].expr, 1 - m.z[4] <= 2 - (m.z[2] + m.z[3])
        )

        assertExpressionsEqual(self, m.cons[10].expr, m.z[4] >= 1)

    def test_xor(self):
        m = self.make_model()
        e = m.a.xor(m.b)

        visitor = LogicalToDisjunctiveVisitor()
        m.cons = visitor.constraints
        m.z = visitor.z_vars
        m.disjuncts = visitor.disjuncts
        m.disjunctions = visitor.disjunctions

        visitor.walk_expression(e)
        self.assertIs(m.a.get_associated_binary(), m.z[1])
        self.assertIs(m.b.get_associated_binary(), m.z[2])

        self.assertEqual(len(m.z), 2)
        self.assertEqual(len(m.cons), 1)
        self.assertEqual(len(m.disjuncts), 2)
        self.assertEqual(len(m.disjunctions), 1)

        assertExpressionsEqual(
            self, m.disjuncts[0].constraint.expr, m.z[1] + m.z[2] == 1
        )
        assertExpressionsEqual(
            self,
            m.disjuncts[1].disjunction.disjuncts[0].constraint[1].expr,
            m.z[1] + m.z[2] <= 0,
        )
        assertExpressionsEqual(
            self,
            m.disjuncts[1].disjunction.disjuncts[1].constraint[1].expr,
            m.z[1] + m.z[2] >= 2,
        )

        assertExpressionsEqual(
            self, m.cons[1].expr, m.disjuncts[0].binary_indicator_var >= 1
        )

    def test_at_most(self):
        m = self.make_model()
        e = atmost(2, m.a, (m.a.land(m.b)), m.c)

        visitor = LogicalToDisjunctiveVisitor()
        m.cons = visitor.constraints
        m.z = visitor.z_vars
        m.disjuncts = visitor.disjuncts
        m.disjunctions = visitor.disjunctions

        visitor.walk_expression(e)
        self.assertIs(m.a.get_associated_binary(), m.z[1])
        a = m.z[1]
        self.assertIs(m.b.get_associated_binary(), m.z[2])
        b = m.z[2]
        self.assertIs(m.c.get_associated_binary(), m.z[4])
        c = m.z[4]

        self.assertEqual(len(m.z), 4)
        self.assertEqual(len(m.cons), 4)
        self.assertEqual(len(m.disjuncts), 2)
        self.assertEqual(len(m.disjunctions), 1)

        # z3 = a ^ b
        assertExpressionsEqual(self, m.cons[1].expr, m.z[3] <= a)
        assertExpressionsEqual(self, m.cons[2].expr, m.z[3] <= b)
        assertExpressionsEqual(self, m.cons[3].expr, 1 - m.z[3] <= 2 - sum([a, b]))

        # atmost in disjunctive form
        assertExpressionsEqual(
            self, m.disjuncts[0].constraint.expr, m.z[1] + m.z[3] + m.z[4] <= 2
        )
        assertExpressionsEqual(
            self, m.disjuncts[1].constraint.expr, m.z[1] + m.z[3] + m.z[4] >= 3
        )
        assertExpressionsEqual(
            self, m.cons[4].expr, m.disjuncts[0].binary_indicator_var >= 1
        )

    def test_at_least(self):
        m = self.make_model()
        e = atleast(2, m.a, m.b, m.c)
        visitor = LogicalToDisjunctiveVisitor()
        m.cons = visitor.constraints
        m.z = visitor.z_vars
        m.disjuncts = visitor.disjuncts
        m.disjunctions = visitor.disjunctions

        visitor.walk_expression(e)
        self.assertIs(m.a.get_associated_binary(), m.z[1])
        a = m.z[1]
        self.assertIs(m.b.get_associated_binary(), m.z[2])
        b = m.z[2]
        self.assertIs(m.c.get_associated_binary(), m.z[3])
        c = m.z[3]

        self.assertEqual(len(m.z), 3)
        self.assertEqual(len(m.cons), 1)

        # atleast in disjunctive form
        assertExpressionsEqual(
            self, m.disjuncts[0].constraint.expr, m.z[1] + m.z[2] + m.z[3] >= 2
        )
        assertExpressionsEqual(
            self, m.disjuncts[1].constraint.expr, m.z[1] + m.z[2] + m.z[3] <= 1
        )

        assertExpressionsEqual(
            self, m.cons[1].expr, m.disjuncts[0].binary_indicator_var >= 1
        )

    @unittest.skipUnless(gurobi_available, "Gurobi is not available")
    def test_logical_integration(self):
        """
        This is kind of a ridiculous test, but I bothered type it and it has
        a lot of logical things together, so adding it.
        """
        m = self.make_model()
        m.d = BooleanVar()
        m.t = BooleanVar()
        e = m.t.equivalent_to(lnot(lor(m.a, m.b)).land(exactly(1, [m.c, m.d])))

        # We're forcing t to be True.
        m.c.fix(True)
        m.d.fix(False)
        m.a.fix(False)
        m.b.fix(False)

        visitor = LogicalToDisjunctiveVisitor()
        m.cons = visitor.constraints
        m.z = visitor.z_vars
        m.disjuncts = visitor.disjuncts
        m.disjunctions = visitor.disjunctions

        visitor.walk_expression(e)

        self.assertEqual(len(m.z), 11)
        self.assertIs(m.a.get_associated_binary(), m.z[2])
        a = m.z[2]
        self.assertIs(m.b.get_associated_binary(), m.z[3])
        b = m.z[3]
        # apologies to the universe for this one, but because my own notation
        # is awful:
        z3 = m.z[4]
        z4 = m.z[5]
        z5 = m.z[8]
        self.assertIs(m.t.get_associated_binary(), m.z[1])
        t = m.z[1]
        z6 = m.z[9]
        z7 = m.z[10]
        self.assertIs(m.c.get_associated_binary(), m.z[6])
        c = m.z[6]
        self.assertIs(m.d.get_associated_binary(), m.z[7])
        d = m.z[7]
        z8 = m.z[11]

        self.assertEqual(len(m.disjuncts), 2)
        self.assertEqual(
            len(
                list(
                    m.disjuncts[0].component_data_objects(
                        Constraint, descend_into=False
                    )
                )
            ),
            1,
        )
        assertExpressionsEqual(self, m.disjuncts[0].constraint.expr, c + d == 1)
        # not nested
        self.assertEqual(
            len(
                list(
                    m.disjuncts[0].component_data_objects(Disjunct, descend_into=False)
                )
            ),
            0,
        )

        # nested
        self.assertEqual(
            len(
                list(
                    m.disjuncts[1].component_data_objects(
                        Constraint, descend_into=False
                    )
                )
            ),
            0,
        )
        self.assertEqual(
            len(
                list(
                    m.disjuncts[1].component_data_objects(Disjunct, descend_into=False)
                )
            ),
            2,
        )
        self.assertEqual(len(m.disjuncts[1].disjunction.disjuncts), 2)
        assertExpressionsEqual(
            self, m.disjuncts[1].disjunction.disjuncts[0].constraint[1].expr, c + d <= 0
        )
        assertExpressionsEqual(
            self, m.disjuncts[1].disjunction.disjuncts[1].constraint[1].expr, c + d >= 2
        )

        self.assertEqual(len(m.disjunctions), 1)
        self.assertIs(m.disjunctions[0].disjuncts[0], m.disjuncts[0])
        self.assertIs(m.disjunctions[0].disjuncts[1], m.disjuncts[1])

        self.assertEqual(len(m.cons), 17)
        assertExpressionsEqual(self, m.cons[1].expr, (1 - z3) + a + b >= 1)
        assertExpressionsEqual(self, m.cons[2].expr, z3 + (1 - a) >= 1)
        assertExpressionsEqual(self, m.cons[3].expr, z3 + (1 - b) >= 1)
        assertExpressionsEqual(self, m.cons[4].expr, z4 == 1 - z3)
        assertExpressionsEqual(self, m.cons[5].expr, z5 <= z4)
        assertExpressionsEqual(
            self, m.cons[6].expr, z5 <= m.disjuncts[0].binary_indicator_var
        )
        assertExpressionsEqual(
            self,
            m.cons[7].expr,
            1 - z5 <= 2 - (z4 + m.disjuncts[0].binary_indicator_var),
        )
        assertExpressionsEqual(self, m.cons[8].expr, (1 - z6) + (1 - t) + z5 >= 1)
        assertExpressionsEqual(self, m.cons[9].expr, 1 - (1 - t) + z6 >= 1)
        assertExpressionsEqual(self, m.cons[10].expr, z6 + (1 - z5) >= 1)
        assertExpressionsEqual(self, m.cons[11].expr, (1 - z7) + (1 - z5) + t >= 1)
        assertExpressionsEqual(self, m.cons[12].expr, z7 + (1 - t) >= 1)
        assertExpressionsEqual(self, m.cons[13].expr, 1 - (1 - z5) + z7 >= 1)
        assertExpressionsEqual(self, m.cons[14].expr, z8 <= z6)
        assertExpressionsEqual(self, m.cons[15].expr, z8 <= z7)
        assertExpressionsEqual(self, m.cons[16].expr, 1 - z8 <= 2 - (z6 + z7))
        assertExpressionsEqual(self, m.cons[17].expr, z8 >= 1)

        TransformationFactory('gdp.bigm').apply_to(m)
        m.obj = Objective(expr=m.t.get_associated_binary())
        SolverFactory('gurobi').solve(m, tee=True)
        update_boolean_vars_from_binary(m)

        self.assertTrue(value(e))
        self.assertEqual(value(m.obj), 1)
        self.assertTrue(value(m.t))

    def test_boolean_fixed_true(self):
        m = self.make_model()
        e = m.a.implies(m.b)
        m.a.fix(True)

        visitor = LogicalToDisjunctiveVisitor()
        m.cons = visitor.constraints
        m.z = visitor.z_vars
        m.disjuncts = visitor.disjuncts
        m.disjunctions = visitor.disjunctions

        visitor.walk_expression(e)
        # we'll get !a v b
        self.assertEqual(len(m.z), 3)
        self.assertEqual(len(m.cons), 4)

        self.assertIs(m.a.get_associated_binary(), m.z[1])
        self.assertTrue(m.z[1].fixed)
        self.assertEqual(value(m.z[1]), 1)
        self.assertIs(m.b.get_associated_binary(), m.z[2])

        assertExpressionsEqual(
            self, m.cons[1].expr, (1 - m.z[3]) + (1 - m.z[1]) + m.z[2] >= 1
        )
        assertExpressionsEqual(self, m.cons[2].expr, 1 - (1 - m.z[1]) + m.z[3] >= 1)
        assertExpressionsEqual(self, m.cons[3].expr, m.z[3] + (1 - m.z[2]) >= 1)
        assertExpressionsEqual(self, m.cons[4].expr, m.z[3] >= 1)

    def test_boolean_fixed_false(self):
        m = self.make_model()
        e = m.a & m.b
        m.a.fix(False)

        visitor = LogicalToDisjunctiveVisitor()
        m.cons = visitor.constraints
        m.z = visitor.z_vars
        m.disjuncts = visitor.disjuncts
        m.disjunctions = visitor.disjunctions

        visitor.walk_expression(e)
        # we'll get !a v b
        self.assertEqual(len(m.z), 3)
        self.assertEqual(len(m.cons), 4)

        self.assertIs(m.a.get_associated_binary(), m.z[1])
        self.assertTrue(m.z[1].fixed)
        self.assertEqual(value(m.z[1]), 0)
        self.assertIs(m.b.get_associated_binary(), m.z[2])

        assertExpressionsEqual(self, m.cons[1].expr, m.z[1] >= m.z[3])
        assertExpressionsEqual(self, m.cons[2].expr, m.z[2] >= m.z[3])
        assertExpressionsEqual(
            self, m.cons[3].expr, 1 - m.z[3] <= 2 - (m.z[1] + m.z[2])
        )
        assertExpressionsEqual(self, m.cons[4].expr, m.z[3] >= 1)

    def test_boolean_fixed_none(self):
        m = self.make_model()
        e = m.a & m.b
        # I don't get what this means, but you can do it, so... I guess we need
        # to handle it.
        m.a.fix(None)

        visitor = LogicalToDisjunctiveVisitor()
        m.cons = visitor.constraints
        m.z = visitor.z_vars
        m.disjuncts = visitor.disjuncts
        m.disjunctions = visitor.disjunctions

        visitor.walk_expression(e)
        # we'll get !a v b
        self.assertEqual(len(m.z), 3)
        self.assertEqual(len(m.cons), 4)

        self.assertIs(m.a.get_associated_binary(), m.z[1])
        self.assertTrue(m.z[1].fixed)
        self.assertIsNone(m.z[1].value)
        self.assertIs(m.b.get_associated_binary(), m.z[2])

        assertExpressionsEqual(self, m.cons[1].expr, m.z[1] >= m.z[3])
        assertExpressionsEqual(self, m.cons[2].expr, m.z[2] >= m.z[3])
        assertExpressionsEqual(
            self, m.cons[3].expr, 1 - m.z[3] <= 2 - (m.z[1] + m.z[2])
        )
        assertExpressionsEqual(self, m.cons[4].expr, m.z[3] >= 1)

    def test_no_need_to_walk(self):
        m = self.make_model()
        e = m.a

        visitor = LogicalToDisjunctiveVisitor()
        m.cons = visitor.constraints
        m.z = visitor.z_vars

        visitor.walk_expression(e)
        self.assertEqual(len(m.z), 1)
        self.assertIs(m.a.get_associated_binary(), m.z[1])
        self.assertEqual(len(m.cons), 1)
        assertExpressionsEqual(self, m.cons[1].expr, m.z[1] >= 1)

    def test_binary_already_associated(self):
        m = self.make_model()
        m.mine = Var(domain=Binary)
        m.a.associate_binary_var(m.mine)

        e = m.a.land(m.b)

        visitor = LogicalToDisjunctiveVisitor()
        m.cons = visitor.constraints
        m.z = visitor.z_vars

        visitor.walk_expression(e)

        self.assertEqual(len(m.z), 2)
        self.assertIs(m.b.get_associated_binary(), m.z[1])
        self.assertEqual(len(m.cons), 4)
        assertExpressionsEqual(self, m.cons[1].expr, m.z[2] <= m.mine)
        assertExpressionsEqual(self, m.cons[2].expr, m.z[2] <= m.z[1])
        assertExpressionsEqual(
            self, m.cons[3].expr, 1 - m.z[2] <= 2 - (m.mine + m.z[1])
        )
        assertExpressionsEqual(self, m.cons[4].expr, m.z[2] >= 1)

    # [ESJ 11/22]: We'll probably eventually support all of these examples, but
    # for now test that we handle them gracefully:
    def test_integer_var_in_at_least(self):
        m = self.make_model()
        m.x = Var(bounds=(0, 10), domain=Integers)
        e = atleast(m.x, m.a, m.b, m.c)

        visitor = LogicalToDisjunctiveVisitor()
        m.cons = visitor.constraints
        m.z = visitor.z_vars

        with self.assertRaisesRegex(
            MouseTrap,
            r"The first argument 'x' to "
            r"'atleast\(x: \[a, b, c\]\)' is potentially variable. "
            r"This may be a mathematically coherent expression; However "
            r"it is not yet supported to convert it to a disjunctive "
            r"program.",
            normalize_whitespace=True,
        ):
            visitor.walk_expression(e)

    def test_numeric_expression_in_at_most(self):
        m = self.make_model()
        m.x = Var([1, 2], bounds=(0, 10), domain=Integers)
        m.y = Var(domain=Integers)
        m.e = Expression(expr=m.x[1] * m.x[2])
        e = atmost(m.e + m.y, m.a, m.b, m.c)

        visitor = LogicalToDisjunctiveVisitor()
        m.cons = visitor.constraints
        m.z = visitor.z_vars

        with self.assertRaisesRegex(
            MouseTrap,
            r"The first argument '\(x\[1\]\*x\[2\]\) \+ y' to "
            r"'atmost\(\(x\[1\]\*x\[2\]\) \+ y: \[a, b, c\]\)' is "
            r"potentially variable. "
            r"This may be a mathematically coherent expression; However "
            r"it is not yet supported to convert it to a disjunctive "
            r"program",
            normalize_whitespace=True,
        ):
            visitor.walk_expression(e)

    def test_named_expression_in_at_most(self):
        m = self.make_model()
        m.x = Var([1, 2], bounds=(0, 10), domain=Integers)
        m.y = Var(domain=Integers)
        m.e = Expression(expr=m.x[1] * m.x[2])
        e = atmost(m.e, m.a, m.b, m.c)

        visitor = LogicalToDisjunctiveVisitor()
        m.cons = visitor.constraints
        m.z = visitor.z_vars

        with self.assertRaisesRegex(
            MouseTrap,
            r"The first argument 'e' to "
            r"'atmost\(\(x\[1\]\*x\[2\]\): \[a, b, c\]\)' is "
            r"potentially variable. "
            r"This may be a mathematically coherent expression; However "
            r"it is not yet supported to convert it to a disjunctive "
            r"program",
            normalize_whitespace=True,
        ):
            visitor.walk_expression(e)

    def test_relational_expr_as_boolean_atom(self):
        m = self.make_model()
        m.x = Var()
        e = m.a.land(m.x >= 3)
        visitor = LogicalToDisjunctiveVisitor()

        with self.assertRaisesRegex(
            MouseTrap,
            "The RelationalExpression '3 <= x' was used as a Boolean "
            "term in a logical proposition. This is not yet supported "
            "when transforming to disjunctive form.",
            normalize_whitespace=True,
        ):
            visitor.walk_expression(e)


class TestLogicalToDisjunctiveTransformation(unittest.TestCase):
    def make_model(self):
        m = ConcreteModel()

        m.a = BooleanVar()
        m.b = BooleanVar([1, 2])
        m.p = Param(initialize=1)
        m.p2 = Param([1, 2], mutable=True)
        m.p2[1] = 1
        m.p2[2] = 2

        m.block = Block()
        m.block.c1 = LogicalConstraint(expr=m.a.land(m.b[1]))
        m.block.c2 = LogicalConstraint(
            expr=exactly(m.p2[2], m.a, m.b[1], m.b[2].lor(m.b[1]))
        )

        m.c1 = LogicalConstraint(expr=atmost(m.p + m.p2[1], m.a, m.b[1], m.b[2]))

        return m

    def check_and_constraints(self, a, b1, z, transBlock):
        assertExpressionsEqual(self, transBlock.transformed_constraints[1].expr, z <= a)
        assertExpressionsEqual(
            self, transBlock.transformed_constraints[2].expr, z <= b1
        )
        assertExpressionsEqual(
            self, transBlock.transformed_constraints[3].expr, 1 - z <= 2 - (a + b1)
        )
        assertExpressionsEqual(self, transBlock.transformed_constraints[4].expr, z >= 1)

    def check_block_c1_transformed(self, m, transBlock):
        self.assertFalse(m.block.c1.active)
        self.assertIs(m.a.get_associated_binary(), transBlock.auxiliary_vars[1])
        self.assertIs(m.b[1].get_associated_binary(), transBlock.auxiliary_vars[2])
        self.check_and_constraints(
            transBlock.auxiliary_vars[1],
            transBlock.auxiliary_vars[2],
            transBlock.auxiliary_vars[3],
            transBlock,
        )

    def check_block_exactly(self, a, b1, b2, z4, transBlock):
        m = transBlock.model()

        # z[4] = b[2] v b[1]
        assertExpressionsEqual(
            self, transBlock.transformed_constraints[5].expr, (1 - z4) + b2 + b1 >= 1
        )
        assertExpressionsEqual(
            self, transBlock.transformed_constraints[6].expr, z4 + (1 - b2) >= 1
        )
        assertExpressionsEqual(
            self, transBlock.transformed_constraints[7].expr, z4 + (1 - b1) >= 1
        )

        # exactly in disjunctive form
        assertExpressionsEqual(
            self,
            transBlock.auxiliary_disjuncts[0].constraint.expr,
            a + b1 + z4 == m.p2[2],
        )
        assertExpressionsEqual(
            self,
            transBlock.auxiliary_disjuncts[1]
            .disjunction.disjuncts[0]
            .constraint[1]
            .expr,
            a + b1 + z4 <= m.p2[2] - 1,
        )
        assertExpressionsEqual(
            self,
            transBlock.auxiliary_disjuncts[1]
            .disjunction.disjuncts[1]
            .constraint[1]
            .expr,
            a + b1 + z4 >= m.p2[2] + 1,
        )

        assertExpressionsEqual(
            self,
            transBlock.transformed_constraints[8].expr,
            transBlock.auxiliary_disjuncts[0].binary_indicator_var >= 1,
        )

    def check_block_transformed(self, m):
        self.assertFalse(m.block.c2.active)
        transBlock = m.block._logical_to_disjunctive
        self.assertEqual(len(transBlock.auxiliary_vars), 5)
        self.assertEqual(len(transBlock.transformed_constraints), 8)
        self.assertEqual(len(transBlock.auxiliary_disjuncts), 2)
        self.assertEqual(len(transBlock.auxiliary_disjunctions), 1)

        self.check_block_c1_transformed(m, transBlock)

        self.assertIs(m.b[2].get_associated_binary(), transBlock.auxiliary_vars[4])

        z4 = transBlock.auxiliary_vars[5]
        a = transBlock.auxiliary_vars[1]
        b1 = transBlock.auxiliary_vars[2]
        b2 = transBlock.auxiliary_vars[4]
        self.check_block_exactly(a, b1, b2, z4, transBlock)

    def test_constraint_target(self):
        m = self.make_model()
        TransformationFactory('contrib.logical_to_disjunctive').apply_to(
            m, targets=[m.block.c1]
        )

        transBlock = m.block._logical_to_disjunctive
        self.assertEqual(len(transBlock.auxiliary_vars), 3)
        self.assertEqual(len(transBlock.transformed_constraints), 4)
        self.assertEqual(len(transBlock.auxiliary_disjuncts), 0)
        self.assertEqual(len(transBlock.auxiliary_disjunctions), 0)
        self.check_block_c1_transformed(m, transBlock)
        self.assertTrue(m.block.c2.active)
        self.assertTrue(m.c1.active)

    def test_block_target(self):
        m = self.make_model()
        TransformationFactory('contrib.logical_to_disjunctive').apply_to(
            m, targets=[m.block]
        )

        self.check_block_transformed(m)
        self.assertTrue(m.c1.active)

    def test_transform_block(self):
        m = self.make_model()
        TransformationFactory('contrib.logical_to_disjunctive').apply_to(m.block)

        self.check_block_transformed(m)
        self.assertTrue(m.c1.active)

    def test_transform_model(self):
        m = self.make_model()
        TransformationFactory('contrib.logical_to_disjunctive').apply_to(m)

        # c1 got transformed first
        self.assertFalse(m.c1.active)
        transBlock = m._logical_to_disjunctive
        self.assertEqual(len(transBlock.auxiliary_vars), 3)
        self.assertEqual(len(transBlock.transformed_constraints), 1)
        self.assertEqual(len(transBlock.auxiliary_disjuncts), 2)
        self.assertEqual(len(transBlock.auxiliary_disjunctions), 1)

        a = m._logical_to_disjunctive.auxiliary_vars[1]
        b1 = m._logical_to_disjunctive.auxiliary_vars[2]
        b2 = m._logical_to_disjunctive.auxiliary_vars[3]

        # atmost in disjunctive form
        assertExpressionsEqual(
            self,
            transBlock.auxiliary_disjuncts[0].constraint.expr,
            a + b1 + b2 <= 1 + m.p2[1],
        )
        assertExpressionsEqual(
            self,
            transBlock.auxiliary_disjuncts[1].constraint.expr,
            a + b1 + b2 >= 1 + m.p2[1] + 1,
        )

        assertExpressionsEqual(
            self,
            transBlock.transformed_constraints[1].expr,
            transBlock.auxiliary_disjuncts[0].binary_indicator_var >= 1,
        )

        # and everything on the block is transformed too
        transBlock = m.block._logical_to_disjunctive
        self.assertEqual(len(transBlock.auxiliary_vars), 2)
        self.assertEqual(len(transBlock.transformed_constraints), 8)
        self.assertEqual(len(transBlock.auxiliary_disjuncts), 2)
        self.assertEqual(len(transBlock.auxiliary_disjunctions), 1)
        self.check_and_constraints(a, b1, transBlock.auxiliary_vars[1], transBlock)
        self.check_block_exactly(a, b1, b2, transBlock.auxiliary_vars[2], transBlock)

    @unittest.skipUnless(gurobi_available, "Gurobi is not available")
    def test_reverse_implication_for_land(self):
        m = ConcreteModel()

        m.t = BooleanVar()
        m.a = BooleanVar()
        m.d = BooleanVar()

        m.c = LogicalConstraint(expr=m.t.equivalent_to(m.a.land(m.d)))

        m.a.fix(True)
        m.d.fix(True)

        m.binary = Var(domain=Binary)
        m.t.associate_binary_var(m.binary)

        m.obj = Objective(expr=m.binary)

        TransformationFactory('contrib.logical_to_disjunctive').apply_to(m)
        TransformationFactory('gdp.bigm').apply_to(m)

        SolverFactory('gurobi').solve(m)

        update_boolean_vars_from_binary(m)
        # Should be 1 (we forced it)
        self.assertEqual(value(m.obj), 1)
        self.assertTrue(value(m.t))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
"""
Continuously stirred tank reactor model, based on
pyomo/examples/doc/pyomobook/nonlinear-ch/react_design/ReactorDesign.py
"""

from pyomo.common.dependencies import pandas as pd
import pyomo.environ as pyo
import pyomo.contrib.parmest.parmest as parmest
from pyomo.contrib.parmest.experiment import Experiment


def reactor_design_model():

    # Create the concrete model
    model = pyo.ConcreteModel()

    # Rate constants
    model.k1 = pyo.Param(
        initialize=5.0 / 6.0, within=pyo.PositiveReals, mutable=True
    )  # min^-1
    model.k2 = pyo.Param(
        initialize=5.0 / 3.0, within=pyo.PositiveReals, mutable=True
    )  # min^-1
    model.k3 = pyo.Param(
        initialize=1.0 / 6000.0, within=pyo.PositiveReals, mutable=True
    )  # m^3/(gmol min)

    # Inlet concentration of A, gmol/m^3
    model.caf = pyo.Param(initialize=10000, within=pyo.PositiveReals, mutable=True)

    # Space velocity (flowrate/volume)
    model.sv = pyo.Param(initialize=1.0, within=pyo.PositiveReals, mutable=True)

    # Outlet concentration of each component
    model.ca = pyo.Var(initialize=5000.0, within=pyo.PositiveReals)
    model.cb = pyo.Var(initialize=2000.0, within=pyo.PositiveReals)
    model.cc = pyo.Var(initialize=2000.0, within=pyo.PositiveReals)
    model.cd = pyo.Var(initialize=1000.0, within=pyo.PositiveReals)

    # Objective
    model.obj = pyo.Objective(expr=model.cb, sense=pyo.maximize)

    # Constraints
    model.ca_bal = pyo.Constraint(
        expr=(
            0
            == model.sv * model.caf
            - model.sv * model.ca
            - model.k1 * model.ca
            - 2.0 * model.k3 * model.ca**2.0
        )
    )

    model.cb_bal = pyo.Constraint(
        expr=(0 == -model.sv * model.cb + model.k1 * model.ca - model.k2 * model.cb)
    )

    model.cc_bal = pyo.Constraint(
        expr=(0 == -model.sv * model.cc + model.k2 * model.cb)
    )

    model.cd_bal = pyo.Constraint(
        expr=(0 == -model.sv * model.cd + model.k3 * model.ca**2.0)
    )

    return model


class ReactorDesignExperiment(Experiment):

    def __init__(self, data, experiment_number):
        self.data = data
        self.experiment_number = experiment_number
        self.data_i = data.loc[experiment_number, :]
        self.model = None

    def create_model(self):
        self.model = m = reactor_design_model()
        return m

    def finalize_model(self):
        m = self.model

        # Experiment inputs values
        m.sv = self.data_i['sv']
        m.caf = self.data_i['caf']

        # Experiment output values
        m.ca = self.data_i['ca']
        m.cb = self.data_i['cb']
        m.cc = self.data_i['cc']
        m.cd = self.data_i['cd']

        return m

    def label_model(self):
        m = self.model

        m.experiment_outputs = pyo.Suffix(direction=pyo.Suffix.LOCAL)
        m.experiment_outputs.update(
            [
                (m.ca, self.data_i['ca']),
                (m.cb, self.data_i['cb']),
                (m.cc, self.data_i['cc']),
                (m.cd, self.data_i['cd']),
            ]
        )

        m.unknown_parameters = pyo.Suffix(direction=pyo.Suffix.LOCAL)
        m.unknown_parameters.update(
            (k, pyo.ComponentUID(k)) for k in [m.k1, m.k2, m.k3]
        )

        return m

    def get_labeled_model(self):
        m = self.create_model()
        m = self.finalize_model()
        m = self.label_model()

        return m


def main():

    # For a range of sv values, return ca, cb, cc, and cd
    results = []
    sv_values = [1.0 + v * 0.05 for v in range(1, 20)]
    caf = 10000
    for sv in sv_values:

        # make model
        model = reactor_design_model()

        # add caf, sv
        model.caf = caf
        model.sv = sv

        # solve model
        solver = pyo.SolverFactory("ipopt")
        solver.solve(model)

        # save results
        results.append([sv, caf, model.ca(), model.cb(), model.cc(), model.cd()])

    results = pd.DataFrame(results, columns=["sv", "caf", "ca", "cb", "cc", "cd"])
    print(results)


if __name__ == "__main__":
    main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for pyomo.opt.base.convert
#

import os
from os.path import abspath, dirname

pyomodir = dirname(abspath(__file__)) + os.sep + ".." + os.sep + ".." + os.sep
currdir = dirname(abspath(__file__)) + os.sep

import pyomo.common.unittest as unittest

from pyomo.common.errors import ApplicationError
from pyomo.common.tempfiles import TempfileManager

import pyomo.opt

old_tempdir = TempfileManager.tempdir


class MockArg:
    def __init__(self):
        pass

    def valid_problem_types(self):
        return [pyomo.opt.ProblemFormat.pyomo]

    def write(self, filename="", format=None):
        pass


class MockArg2(MockArg):
    def valid_problem_types(self):
        return [pyomo.opt.ProblemFormat.nl]

    def write(self, filename="", format=None):
        OUTPUT = open(filename, "w")
        INPUT = open(currdir + "test4.nl")
        for line in INPUT:
            print >> OUTPUT, line,
        OUTPUT.close()
        INPUT.close()


class MockArg3(MockArg):
    def valid_problem_types(self):
        return [pyomo.opt.ProblemFormat.mod]

    def write(self, filename="", format=None):
        pass


class MockArg4(MockArg):
    def write(self, filename="", format=None):
        OUTPUT = open(filename, "w")
        INPUT = open(currdir + "test4.nl")
        for line in INPUT:
            print >> OUTPUT, line,
        OUTPUT.close()
        INPUT.close()


class OptConvertDebug(unittest.TestCase):
    def setUp(self):
        TempfileManager.push()
        TempfileManager.tempdir = currdir

    def tearDown(self):
        TempfileManager.clear_tempfiles()
        TempfileManager.tempdir = old_tempdir

    def test_nl_nl1(self):
        """Convert from NL to NL"""
        ans = pyomo.opt.convert_problem(
            ("test4.nl",), None, [pyomo.opt.ProblemFormat.nl]
        )
        self.assertEqual(ans[0], ("test4.nl",))

    def test_nl_nl2(self):
        """Convert from NL to NL"""
        ans = pyomo.opt.convert_problem(
            ("test4.nl", "tmp.nl"), None, [pyomo.opt.ProblemFormat.nl]
        )
        self.assertEqual(ans[0], ("test4.nl", "tmp.nl"))

    def test_error1(self):
        """No valid problem types"""
        try:
            pyomo.opt.convert_problem(
                ("test4.nl", "tmp.nl"), pyomo.opt.ProblemFormat.nl, []
            )
            self.fail("Expected pyomo.opt.ConverterError exception")
        except pyomo.opt.ConverterError:
            pass

    def test_error2(self):
        """Target problem type is not valid"""
        try:
            pyomo.opt.convert_problem(
                ("test4.nl", "tmp.nl"),
                pyomo.opt.ProblemFormat.nl,
                [pyomo.opt.ProblemFormat.mps],
            )
            self.fail("Expected pyomo.opt.ConverterError exception")
        except pyomo.opt.ConverterError:
            pass

    def test_error3(self):
        """Empty argument list"""
        try:
            pyomo.opt.convert_problem((), None, [pyomo.opt.ProblemFormat.mps])
            self.fail("Expected pyomo.opt.ConverterError exception")
        except pyomo.opt.ConverterError:
            pass

    def test_error4(self):
        """Unknown source type"""
        try:
            pyomo.opt.convert_problem(
                ("prob.foo",), None, [pyomo.opt.ProblemFormat.mps]
            )
            self.fail("Expected pyomo.opt.ConverterError exception")
        except pyomo.opt.ConverterError:
            pass

    def test_error5(self):
        """Unknown source type"""
        try:
            pyomo.opt.convert_problem(
                ("prob.lp",), pyomo.opt.ProblemFormat.nl, [pyomo.opt.ProblemFormat.nl]
            )
            self.fail("Expected pyomo.opt.ConverterError exception")
        except pyomo.opt.ConverterError:
            pass

    def test_error6(self):
        """Cannot use pico_convert with more than one file"""
        try:
            ans = pyomo.opt.convert_problem(
                (currdir + "test4.nl", "foo"), None, [pyomo.opt.ProblemFormat.cpxlp]
            )
            self.fail("Expected pyomo.opt.ConverterError exception")
        except pyomo.opt.ConverterError:
            pass

    def test_error8(self):
        """Error when source file cannot be found"""
        try:
            ans = pyomo.opt.convert_problem(
                (currdir + "unknown.nl",), None, [pyomo.opt.ProblemFormat.cpxlp]
            )
            self.fail("Expected pyomo.opt.ConverterError exception")
        except ApplicationError:
            if pyomo.common.Executable("pico_convert"):
                self.fail(
                    "Expected ApplicationError because pico_convert is not available"
                )
            return
        except pyomo.opt.ConverterError:
            pass

    def test_error9(self):
        """The Opt configuration has not been initialized"""
        cmd = pyomo.common.Executable("pico_convert")
        if cmd:
            cmd.disable()
        try:
            ans = pyomo.opt.convert_problem(
                (currdir + "test4.nl",), None, [pyomo.opt.ProblemFormat.cpxlp]
            )
            self.fail("This test didn't fail, but pico_convert should not be defined.")
        except pyomo.opt.ConverterError:
            pass
        cmd.rehash()

    def test_error10(self):
        """GLPSOL can only convert file data"""
        try:
            arg = MockArg3()
            ans = pyomo.opt.convert_problem(
                (arg, pyomo.opt.ProblemFormat.cpxlp, arg),
                None,
                [pyomo.opt.ProblemFormat.cpxlp],
            )
            self.fail("This test didn't fail, but glpsol cannot handle objects.")
        except pyomo.opt.ConverterError:
            pass

    def test_error11(self):
        """Cannot convert MOD that contains data"""
        try:
            ans = pyomo.opt.convert_problem(
                (currdir + "test3.mod", currdir + "test5.dat"),
                None,
                [pyomo.opt.ProblemFormat.cpxlp],
            )
            self.fail(
                "Expected pyomo.opt.ConverterError exception because we provided a MOD file with a 'data;' declaration"
            )
        except ApplicationError:
            if pyomo.common.Executable("glpsol"):
                self.fail("Expected ApplicationError because glpsol is not available")
            return
        except pyomo.opt.ConverterError:
            pass


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import bisect


def find_nearest_index(array, target, tolerance=None):
    # array needs to be sorted and we assume it is zero-indexed
    lo = 0
    hi = len(array)
    i = bisect.bisect_right(array, target, lo=lo, hi=hi)
    # i is the index at which target should be inserted if it is to be
    # right of any equal components.

    if i == lo:
        # target is less than every entry of the set
        nearest_index = i
        delta = array[nearest_index] - target
    elif i == hi:
        # target is greater than or equal to every entry of the set
        nearest_index = i - 1
        delta = target - array[nearest_index]
    else:
        # p_le <= target < p_g
        # delta_left = target - p_le
        # delta_right = p_g - target
        # delta = min(delta_left, delta_right)
        # Tie goes to the index on the left.
        delta, nearest_index = min((abs(target - array[j]), j) for j in [i - 1, i])

    if tolerance is not None:
        if delta > tolerance:
            return None
    return nearest_index


def _distance_from_interval(point, interval, tolerance=None):
    lo, hi = interval
    if tolerance is None:
        tolerance = 0.0
    if point < lo - tolerance:
        return lo - point
    elif lo - tolerance <= point and point <= hi + tolerance:
        return 0.0
    elif point > hi + tolerance:
        return point - hi


def find_nearest_interval_index(
    interval_array, target, tolerance=None, prefer_left=True
):
    # NOTE: This function quickly begins to behave badly if tolerance
    # gets too large, e.g. greater than the width of the smallest
    # interval. For this reason, intervals that represent a single
    # point, e.g. (1.0, 1.0) should not be supported.
    array_lo = 0
    array_hi = len(interval_array)
    target_tuple = (target,)
    i = bisect.bisect_right(interval_array, target_tuple, lo=array_lo, hi=array_hi)
    distance_tol = 0.0 if tolerance is None else tolerance
    if i == array_lo:
        # We are at or to the left of the left endpoint of the
        # first interval.
        nearest_index = i
        delta = _distance_from_interval(
            target, interval_array[i], tolerance=distance_tol
        )
    elif i == array_hi:
        # We are within or to the right of the last interval.
        nearest_index = i - 1
        delta = _distance_from_interval(
            target, interval_array[i - 1], tolerance=distance_tol
        )
    else:
        # Find closest interval
        if prefer_left:
            # In the case of a tie, we return the left interval
            # by default.
            delta, nearest_index = min(
                (
                    _distance_from_interval(
                        target, interval_array[j], tolerance=distance_tol
                    ),
                    j,
                )
                for j in [i - 1, i]
            )
        else:
            # If prefer_left=False, we return the right interval.
            delta, neg_nearest_index = min(
                (
                    _distance_from_interval(
                        target, interval_array[j], tolerance=distance_tol
                    ),
                    -j,
                )
                for j in [i - 1, i]
            )
            nearest_index = -neg_nearest_index

    # If we have two adjacent intervals, e.g. [(0, 1), (1, 2)], and are just
    # to the right of the boundary, we will not check the left interval as
    # bisect places our tuple, e.g. (1.0+1e-10,), to the right of the right
    # interval.
    if prefer_left and nearest_index >= array_lo + 1:
        delta_left = _distance_from_interval(
            target, interval_array[nearest_index - 1], tolerance=distance_tol
        )
        if delta_left <= delta:
            nearest_index = nearest_index - 1
            delta = delta_left
    elif not prefer_left and nearest_index < array_hi - 1:
        delta_right = _distance_from_interval(
            target, interval_array[nearest_index + 1], tolerance=distance_tol
        )
        if delta_right <= delta:
            nearest_index = nearest_index + 1
            delta = delta_right

    if tolerance is not None:
        if delta > tolerance:
            return None
    return nearest_index



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.pyomoobject import PyomoObject
from pyomo.core.base.range import NonNumericRange

GlobalSets = {}


def _get_global_set(name):
    return GlobalSets[name]


_get_global_set.__safe_for_unpickling__ = True


class GlobalSetBase(PyomoObject):
    """The base class for all Global sets"""

    __slots__ = ()

    def __reduce__(self):
        # Cause pickle to preserve references to this object
        return _get_global_set, (self.local_name,)

    def __deepcopy__(self, memo):
        # Prevent deepcopy from duplicating this object
        return self

    def __str__(self):
        # Override str() to always print out the global set name
        return self.name

    #
    # Override the private "_parent" attribute: as this is a
    # _global_ set, we disallow assigning the set to a block.
    #
    @property
    def _parent(self):
        return None

    @_parent.setter
    def _parent(self, val):
        if val is None:
            return
        val = val()  # dereference the weakref
        raise RuntimeError(
            "Cannot assign a GlobalSet '%s' to %s '%s'"
            % (
                self.global_name,
                'model' if val.model() is val else 'block',
                val.name or 'unknown',
            )
        )


# FIXME: This mocks up part of the Set API until we can break up the set
# module to resolve circular dependencies and can make this a proper
# GlobalSet (Scalar IndexedComponent objects are indexed by
# UnindexedComponent_set, but we would like UnindexedComponent_set to be
# a proper scalar IndexedComponent).
#
# UnindexedComponent_set = set([None])
class _UnindexedComponent_set(GlobalSetBase):
    local_name = 'UnindexedComponent_set'

    _anonymous_sets = GlobalSetBase

    def __init__(self, name):
        self.name = name
        self._constructed = True

    def __contains__(self, val):
        return val is None

    def get(self, value, default):
        if value is None:
            return value
        return default

    def __iter__(self):
        return (None,).__iter__()

    def __reversed__(self):
        return iter(self)

    def ordered_iter(self):
        return iter(self)

    def sorted_iter(self):
        return iter(self)

    def data(self):
        return tuple(self)

    def ordered_data(self):
        return tuple(self)

    def sorted_data(self):
        return tuple(self)

    def subsets(self, expand_all_set_operators=None):
        return [self]

    def construct(self):
        pass

    def ranges(self):
        yield NonNumericRange(None)

    def bounds(self):
        return (None, None)

    def get_interval(self):
        return (None, None, None)

    def __len__(self):
        return 1

    def __eq__(self, other):
        return self is other

    def __ne__(self, other):
        return self is not other

    @property
    def dimen(self):
        return 0

    def isdiscrete(self):
        return True

    def isfinite(self):
        return True

    def isordered(self):
        # As this set only has a single element, it is implicitly "ordered"
        return True

    def at(self, index):
        if index == 1:
            return None
        raise IndexError("%s index out of range" % (self.name,))

    def ord(self, item):
        if item is None:
            return 1
        raise IndexError(
            "Cannot identify position of %s in Set %s: item not in Set"
            % (item, self.name)
        )

    def first(self):
        return None

    def last(self):
        return None

    def next(self, item, step=1):
        # make sure that item is None (and raise the standard error if it is not)
        self.ord(item)
        if step < 0:
            raise IndexError("Cannot advance before the beginning of the Set")
        else:
            raise IndexError("Cannot advance past the end of the Set")

    def nextw(self, item, step=1):
        # make sure that item is None (and raise the standard error if it is not)
        self.ord(item)
        return None

    def prev(self, item, step=1):
        return self.next(item, -step)

    def prevw(self, item, step=1):
        return self.nextw(item, -step)

    def parent_block(self):
        return None

    def parent_component(self):
        return self


UnindexedComponent_set = _UnindexedComponent_set('UnindexedComponent_set')
GlobalSets[UnindexedComponent_set.local_name] = UnindexedComponent_set

UnindexedComponent_index = next(iter(UnindexedComponent_set))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Helper functions for variable reinitialization."""

import logging
import random

from pyomo.core import Var

logger = logging.getLogger('pyomo.contrib.multistart')


def rand(val, lb, ub):
    return random.uniform(lb, ub)  # uniform distribution between lb and ub


def midpoint_guess_and_bound(val, lb, ub):
    """Midpoint between current value and farthest bound."""
    far_bound = ub if ((ub - val) >= (val - lb)) else lb  # farther bound
    return (far_bound + val) / 2


def rand_guess_and_bound(val, lb, ub):
    """Random choice between current value and farthest bound."""
    far_bound = ub if ((ub - val) >= (val - lb)) else lb  # farther bound
    return random.uniform(val, far_bound)


def rand_distributed(val, lb, ub, divisions=9):
    """Random choice among evenly distributed set of values between bounds."""
    set_distributed_vals = linspace(lb, ub, divisions)
    return random.choice(set_distributed_vals)


def simple_midpoint(val, lb, ub):
    return (lb + ub) * 0.5


def linspace(lower, upper, n):
    """Linearly spaced range."""
    return [lower + x * (upper - lower) / (n - 1) for x in range(n)]


strategies = {
    "rand": rand,
    "midpoint_guess_and_bound": midpoint_guess_and_bound,
    "rand_guess_and_bound": rand_guess_and_bound,
    "rand_distributed": rand_distributed,
    "midpoint": simple_midpoint,
}


def reinitialize_variables(model, config):
    """Reinitializes all variable values in the model.

    Excludes fixed, noncontinuous, and unbounded variables.

    """
    for var in model.component_data_objects(ctype=Var, descend_into=True):
        if var.is_fixed() or not var.is_continuous():
            continue
        if var.lb is None or var.ub is None:
            if not config.suppress_unbounded_warning:
                logger.warning(
                    'Skipping reinitialization of unbounded variable '
                    '%s with bounds (%s, %s). '
                    'To suppress this message, set the '
                    'suppress_unbounded_warning flag.' % (var.name, var.lb, var.ub)
                )
            continue
        val = var.value if var.value is not None else (var.lb + var.ub) / 2
        # apply reinitialization strategy to variable
        var.set_value(
            strategies[config.strategy](val, var.lb, var.ub), skip_validation=True
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common import DeveloperError
from pyomo.common.collections import (
    UserDict,
    Mapping,
    MutableMapping,
    Set as collections_Set,
    Sequence,
)
from pyomo.common.modeling import NOTSET
from pyomo.core.base.set import DeclareGlobalSet, Set, SetOf, OrderedSetOf, SetData
from pyomo.core.base.component import Component, ComponentData
from pyomo.core.base.global_set import UnindexedComponent_set
from pyomo.core.base.enums import SortComponents
from pyomo.core.base.indexed_component import IndexedComponent, normalize_index
from pyomo.core.base.indexed_component_slice import (
    IndexedComponent_slice,
    _IndexedComponent_slice_iter,
)
from pyomo.core.base.util import flatten_tuple
from pyomo.common.deprecation import deprecated

_UnindexedComponent_key = list(UnindexedComponent_set)
_UnindexedComponent_base_key = tuple(UnindexedComponent_set)

DeclareGlobalSet(
    Set(
        initialize=UnindexedComponent_set,
        name='UnindexedComponent_ReferenceSet',
        doc='An indexing set used by references to unindexed (scalar) '
        'components that is equivalent to but NOT the UnindexedComponent_set',
    ),
    globals(),
)


class _fill_in_known_wildcards:
    """Variant of "six.advance_iterator" that substitutes wildcard values

    This object is initialized with a tuple of index values.  Calling
    the resulting object on a :py:class:`_slice_generator` will
    "advance" the iterator, substituting values from the tuple into the
    slice wildcards (":" indices), and returning the resulting object.
    The motivation for implementing this as an iterator is so that we
    can re-use all the logic from
    :py:meth:`_IndexedComponent_slice_iter.__next__` when looking up
    specific indices within the slice.

    Parameters
    ----------
    wildcard_values : tuple of index values
        a tuple containing index values to substitute into the slice wildcards

    look_in_index : :py:class:`bool` [optional]
        If True, the iterator will also look for matches using the
        components' underlying index_set() in addition to the (sparse)
        indices matched by the components' __contains__()
        method. [default: False]

    get_if_not_present : :py:class:`bool` [optional]
        If True, the iterator will attempt to retrieve data objects
        (through getitem) for indexes that match the underlying
        component index_set() but do not appear in the (sparse) indices
        matched by __contains__.  get_if_not_present implies
        look_in_index.  [default: False]

    """

    def __init__(self, wildcard_values, look_in_index=False, get_if_not_present=False):
        self.base_key = wildcard_values
        self.key = list(wildcard_values)
        self.known_slices = set()
        self.look_in_index = look_in_index or get_if_not_present
        self.get_if_not_present = get_if_not_present

    def __call__(self, _slice):
        """Advance the specified slice generator, substituting wildcard values

        This advances the passed :py:class:`_slice_generator
        <pyomo.core.base.indexed_component_slice._slice_generator>` by
        substituting values from the `wildcard_values` list for any
        wildcard slices ("`:`").

        Parameters
        ----------
        _slice : pyomo.core.base.indexed_component_slice._slice_generator
            the slice to advance
        """
        if _slice in self.known_slices:
            # We have encountered the same slice twice, i.e. IC_slice_iter
            # has constructed an iter_stack, and is now attempting to
            # exhaust the most recent iterator in that stack.
            # We know then that the iter stack is complete, and we
            # tell IC_slice_iter.__next__ to wrap up by raising
            # StopIteration.
            raise StopIteration()
        self.known_slices.add(_slice)

        # `idx_count` is the number of indices this component needs.
        if _slice.ellipsis is None:
            idx_count = _slice.explicit_index_count
        elif not _slice.component.is_indexed():
            idx_count = 1
        else:
            idx_count = _slice.component.index_set().dimen
            if idx_count is None:
                raise SliceEllipsisLookupError(
                    "Cannot lookup elements in a _ReferenceDict when the "
                    "underlying slice object contains ellipsis over a jagged "
                    "(dimen=None) Set"
                )
        try:
            # Here we assemble the index we will actually use to access
            # the component.
            idx = tuple(
                _slice.fixed[i] if i in _slice.fixed else self.key.pop(0)
                for i in range(idx_count)
            )
            # _slice corresponds to some sliced entry in the call/iter stacks
            # that contains the information describing the slice.
            # Here we fill in an index with the fixed indices from the slice
            # and the wildcard indices from the provided key.
        except IndexError:
            # Occurs if we try to pop from an empty list.
            raise KeyError(
                "Insufficient values for slice of indexed component '%s' "
                "(found evaluating slice index %s)"
                % (_slice.component.name, self.base_key)
            )

        if idx in _slice.component:
            # We have found a matching component at this level of the
            # underlying slice _call_stack.  We need to record the index
            # for later use (i.e., with get_last_index() /
            # get_last_index_wildcards()).
            _slice.last_index = idx
            # Return the component. We have successfully accessed
            # the data value.
            return _slice.component[idx]
        elif len(idx) == 1 and idx[0] in _slice.component:
            # `idx` is a len-1 tuple.  It is the scalar, not the tuple,
            # that is contained by the component.
            _slice.last_index = idx
            return _slice.component[idx[0]]
        elif not idx:
            return _slice.component
        elif self.look_in_index:
            # At this point we know our component is sparse and we did
            # not find the component data.  Since `look_in_index` is
            # True, we will verify that the index is valid in the
            # underlying indexing set, and if it is, we will trigger the
            # creation (and return) of the new component data.
            if idx in _slice.component.index_set():
                _slice.last_index = idx
                return _slice.component[idx] if self.get_if_not_present else None
            elif len(idx) == 1 and idx[0] in _slice.component.index_set():
                _slice.last_index = idx
                return _slice.component[idx[0]] if self.get_if_not_present else None

        raise KeyError(
            "Index %s is not valid for indexed component '%s' "
            "(found evaluating slice index %s)"
            % (idx, _slice.component.name, self.base_key)
        )

    def check_complete(self):
        if not self.key:
            return
        if (
            self.key == _UnindexedComponent_key
            and self.base_key == _UnindexedComponent_base_key
        ):
            return
        raise KeyError("Extra (unused) values for slice index %s" % (self.base_key,))


class SliceEllipsisLookupError(LookupError):
    pass


class _ReferenceDict(MutableMapping):
    """A dict-like object whose values are defined by a slice.

    This implements a dict-like object whose keys and values are defined
    by a component slice (:py:class:`IndexedComponent_slice`).  The
    intent behind this object is to replace the normal ``_data``
    :py:class:`dict` in :py:class:`IndexedComponent` containers to
    create "reference" components.

    Parameters
    ----------
    component_slice : :py:class:`IndexedComponent_slice`
        The slice object that defines the "members" of this mutable mapping.
    """

    def __init__(self, component_slice):
        self._slice = component_slice

    def __contains__(self, key):
        try:
            next(self._get_iter(self._slice, key))
            # This calls IC_slice_iter.__next__, which calls
            # _fill_in_known_wildcards.
            return True
        except SliceEllipsisLookupError:
            # We failed because the notion of "location" in an index
            # is ambiguous if the index_set has dimen None.
            # We can still brute-force the lookup by comparing the key
            # to the wildcards of the `last_index` cached by the
            # slice's iterator stack.
            if type(key) is tuple and len(key) == 1:
                key = key[0]
            # Brute force (linear time) lookup
            _iter = iter(self._slice)
            for item in _iter:
                if _iter.get_last_index_wildcards() == key:
                    return True
            return False
        except StopIteration:
            return False
        except LookupError as e:
            if normalize_index.flatten:
                return False
            try:
                next(self._get_iter(self._slice, (key,)))
                return True
            except LookupError:
                return False

    def __getitem__(self, key):
        try:
            # This calls IC_slice_iter.__next__, which calls
            # _fill_in_known_wildcards.
            return next(self._get_iter(self._slice, key, get_if_not_present=True))
        except SliceEllipsisLookupError:
            if type(key) is tuple and len(key) == 1:
                key = key[0]
            # Brute force (linear time) lookup
            _iter = iter(self._slice)
            for item in _iter:
                if _iter.get_last_index_wildcards() == key:
                    return item
            raise KeyError("KeyError: %s" % (key,))
        except (StopIteration, LookupError):
            raise KeyError("KeyError: %s" % (key,))

    def __setitem__(self, key, val):
        tmp = self._slice.duplicate()
        op = tmp._call_stack[-1][0]
        # Replace the end of the duplicated slice's call stack (deepest
        # level of the component hierarchy) with the appropriate
        # `set_item` call.  Then implement the actual __setitem__ by
        # advancing the duplicated iterator.
        if op == IndexedComponent_slice.get_item:
            tmp._call_stack[-1] = (
                IndexedComponent_slice.set_item,
                tmp._call_stack[-1][1],
                val,
            )
        elif op == IndexedComponent_slice.slice_info:
            tmp._call_stack[-1] = (
                IndexedComponent_slice.set_item,
                tmp._call_stack[-1][1],
                val,
            )
        elif op == IndexedComponent_slice.get_attribute:
            tmp._call_stack[-1] = (
                IndexedComponent_slice.set_attribute,
                tmp._call_stack[-1][1],
                val,
            )
        else:
            raise DeveloperError("Unexpected slice _call_stack operation: %s" % op)
        try:
            next(self._get_iter(tmp, key, get_if_not_present=True))
        except StopIteration:
            pass

    def __delitem__(self, key):
        tmp = self._slice.duplicate()
        op = tmp._call_stack[-1][0]
        if op == IndexedComponent_slice.get_item:
            # If the last attribute of the slice gets an item,
            # change it to delete the item
            tmp._call_stack[-1] = (
                IndexedComponent_slice.del_item,
                tmp._call_stack[-1][1],
            )
        elif op == IndexedComponent_slice.slice_info:
            assert len(tmp._call_stack) == 1
            _iter = self._get_iter(tmp, key)
            # The iterator should map all StopIteration exceptions to KeyErrors
            next(_iter)
            del _iter._iter_stack[0].component[_iter.get_last_index()]
            return
        elif op == IndexedComponent_slice.get_attribute:
            # If the last attribute of the slice retrieves an attribute,
            # change it to delete the attribute
            tmp._call_stack[-1] = (
                IndexedComponent_slice.del_attribute,
                tmp._call_stack[-1][1],
            )
        else:
            raise DeveloperError("Unexpected slice _call_stack operation: %s" % op)
        try:
            next(self._get_iter(tmp, key))
        except StopIteration:
            pass

    def __iter__(self):
        return self._slice.wildcard_keys(SortComponents.UNSORTED)

    def __len__(self):
        # Note that unlike for regular dicts, len() of a _ReferenceDict
        # is very slow (linear time).
        return sum(1 for i in self._slice)

    def keys(self, sort=SortComponents.UNSORTED):
        return self._slice.wildcard_keys(sort)

    def items(self, sort=SortComponents.UNSORTED):
        """Return the wildcard, value tuples for this ReferenceDict

        This method is necessary because the default implementation
        iterates over the keys and looks the values up in the
        dictionary.  Unfortunately some slices have structures that make
        looking up components by the wildcard keys very expensive
        (linear time; e.g., the use of ellipses with jagged sets).  By
        implementing this method without using lookups, general methods
        that iterate over everything (like component.pprint()) will
        still be linear and not quadratic time.

        """
        return self._slice.wildcard_items(sort)

    def values(self, sort=SortComponents.UNSORTED):
        """Return the values for this ReferenceDict

        This method is necessary because the default implementation
        iterates over the keys and looks the values up in the
        dictionary.  Unfortunately some slices have structures that make
        looking up components by the wildcard keys very expensive
        (linear time; e.g., the use of ellipses with jagged sets).  By
        implementing this method without using lookups, general methods
        that iterate over everything (like component.pprint()) will
        still be linear and not quadratic time.

        """
        return self._slice.wildcard_values(sort)

    @deprecated('The iteritems method is deprecated. Use dict.items().', version='6.0')
    def iteritems(self):
        return self.items()

    @deprecated(
        'The itervalues method is deprecated. Use dict.values().', version='6.0'
    )
    def itervalues(self):
        return self.values()

    def _get_iter(self, _slice, key, get_if_not_present=False):
        # Construct a slice iter with `_fill_in_known_wildcards`
        # as `advance_iter`. This reuses all the logic from the slice
        # iter to walk the call/iter stacks, but just jumps to a
        # particular component rather than actually iterating.
        # This is how this object does lookups.
        if key.__class__ not in (tuple, list):
            key = (key,)
        if normalize_index.flatten:
            key = flatten_tuple(key)
        return _IndexedComponent_slice_iter(
            _slice, _fill_in_known_wildcards(key, get_if_not_present=get_if_not_present)
        )


class _ReferenceDict_mapping(UserDict):
    def __init__(self, data):
        # Overwrite the internal dict to keep a reference to whatever
        # mapping was passed in (potentially preserving the
        # "orderedness" of the source dictionary).
        self.data = data


class _ReferenceSet(collections_Set):
    """A set-like object whose values are defined by a slice.

    This implements a set-like object whose members are defined by a
    component slice (:py:class:`IndexedComponent_slice`).
    :py:class:`_ReferenceSet` differs from the
    :py:class:`_ReferenceDict` above in that it looks in the underlying
    component ``index_set()`` for values that match the slice, and not
    just the (sparse) indices defined by the slice.

    Parameters
    ----------
    component_slice : :py:class:`IndexedComponent_slice`
        The slice object that defines the "members" of this set

    """

    def __init__(self, component_slice):
        self._slice = component_slice

    def __contains__(self, key):
        try:
            next(self._get_iter(self._slice, key))
            return True
        except SliceEllipsisLookupError:
            if type(key) is tuple and len(key) == 1:
                key = key[0]
            # Brute force (linear time) lookup
            _iter = iter(self._slice)
            for item in _iter:
                if _iter.get_last_index_wildcards() == key:
                    return True
            return False
        except (StopIteration, LookupError):
            return False

    def __iter__(self):
        return self._slice.index_wildcard_keys(False)

    def __len__(self):
        return sum(1 for _ in self)

    def _get_iter(self, _slice, key):
        if key.__class__ not in (tuple, list):
            key = (key,)
        if normalize_index.flatten:
            key = flatten_tuple(key)
        return _IndexedComponent_slice_iter(
            _slice,
            _fill_in_known_wildcards(key, look_in_index=True),
            iter_over_index=True,
        )

    def __str__(self):
        return "ReferenceSet(%s)" % (self._slice,)

    def ordered_iter(self):
        return self._slice.index_wildcard_keys(SortComponents.ORDERED_INDICES)

    def sorted_iter(self):
        return self._slice.index_wildcard_keys(SortComponents.SORTED_INDICES)


def _identify_wildcard_sets(iter_stack, index):
    # Note that we can only _identify_wildcard_sets for a Reference if
    # normalize_index.flatten is True.
    #
    # If we have already decided that there isn't a common index for the
    # slices, there is nothing more we can do.  Bail.
    if index is None:
        return index

    # Walk the iter_stack that led to the current item and try to
    # identify the component wildcard sets
    #
    # `wildcard_stack` will be a list of dicts mirroring the
    # _iter_stack.  Each dict maps position within that level's
    # component's "subsets" list to the set at that position if it is a
    # wildcard set.
    wildcard_stack = [None] * len(iter_stack)
    for i, level in enumerate(iter_stack):
        if level is not None:
            offset = 0
            # `offset` is the position, within the total index tuple,
            # of the first coordinate of a set.
            wildcard_sets = {}
            # `wildcard_sets` maps position in the current level's
            # "subsets list" to its set if that set is a wildcard.
            for j, s in enumerate(
                level.component.index_set().subsets(expand_all_set_operators=False)
            ):
                # Iterate over the sets that could possibly be wildcards
                if s is UnindexedComponent_set:
                    wildcard_sets[j] = s
                    offset += 1
                    continue
                if s.dimen is None:
                    return None
                # `wildcard_count` is the number of coordinates of this
                # set (which may be multi-dimensional) that have been
                # sliced.
                if level.fixed is None:
                    wildcard_count = s.dimen
                else:
                    wildcard_count = sum(
                        1 for k in range(s.dimen) if k + offset not in level.fixed
                    )
                # `k+offset` is a position in the "total" (flattened)
                # index tuple.  All the _slice_generator's information
                # is in terms of this total index tuple.
                if wildcard_count == s.dimen:
                    # Every coordinate of this subset is covered by a
                    # wildcard. This could happen because of explicit
                    # slices or an ellipsis.
                    wildcard_sets[j] = s
                elif wildcard_count != 0:
                    # This subset is "touched" by an explicit slice, but
                    # the whole set is not (i.e. there is a fixed
                    # component to this subset).  Therefore, as we
                    # cannot extract that subset, we quit.
                    #
                    # We do not simply `continue` as this "partially sliced
                    # set" has ruined our chance of extracting sets that
                    # perfectly describe our wildcards.
                    return None
                offset += s.dimen
            # I believe that this check is not necessary: weirdnesses
            # with ellipsis should get caught by the check for s.dimen
            # above.
            #
            # if offset != level.explicit_index_count:
            #    return None
            wildcard_stack[i] = wildcard_sets
    if not index:
        # index is an empty list, i.e. no wildcard sets have been
        # previously identified.  `wildcard_stack` will serve as the
        # basis for comparison for future components' wildcard sets.
        return wildcard_stack

    # For objects to have "the same" wildcard sets, the same sets must
    # be sliced at the same coordinates of their "subsets list" at the
    # same level of the _iter_stack.

    # Any of the following would preclude identifying common sets
    # among the objects defined by a slice.
    # However, I can't see a way to actually create any of these
    # situations (i.e., I can't test them).  Assertions are left in for
    # defensive programming.
    assert len(index) == len(wildcard_stack)

    for i, level in enumerate(wildcard_stack):
        assert (index[i] is None) == (level is None)
        # No slices at this level in the slice
        if level is None:
            continue
        # if there are a differing number of wildcard "subsets"
        if len(index[i]) != len(level):
            return None
        # if any wildcard "subset" differs in position or set.
        if any(index[i].get(j, None) is not _set for j, _set in level.items()):
            return None
        # These checks seem to intentionally preclude
        #     m.b1[:].v and m.b2[1,:].v
        # from having a common set, even if the sliced set is the same.
        # This is correct, as you cannot express a single reference that
        # covers this example.  You would have to create intermediate
        # References to unify the indexing scheme, e.g.:
        #     m.c[1] = Reference(m.b1[:])
        #     m.c[2] = Reference(m.b2[1,:])
        #     Reference(m.c[:].v)
    return index


def Reference(reference, ctype=NOTSET):
    """Creates a component that references other components

    ``Reference`` generates a *reference component*; that is, an indexed
    component that does not contain data, but instead references data
    stored in other components as defined by a component slice.  The
    ctype parameter sets the :py:meth:`Component.type` of the resulting
    indexed component.  If the ctype parameter is not set and all data
    identified by the slice (at construction time) share a common
    :py:meth:`Component.type`, then that type is assumed.  If either the
    ctype parameter is ``None`` or the data has more than one ctype, the
    resulting indexed component will have a ctype of
    :py:class:`IndexedComponent`.

    If the indices associated with wildcards in the component slice all
    refer to the same :py:class:`Set` objects for all data identified by
    the slice, then the resulting indexed component will be indexed by
    the product of those sets.  However, if all data do not share common
    set objects, or only a subset of indices in a multidimensional set
    appear as wildcards, then the resulting indexed component will be
    indexed by a :py:class:`SetOf` containing a
    :py:class:`_ReferenceSet` for the slice.

    Parameters
    ----------
    reference : :py:class:`IndexedComponent_slice`
        component slice that defines the data to include in the
        Reference component

    ctype : :py:class:`type` [optional]
        the type used to create the resulting indexed component.  If not
        specified, the data's ctype will be used (if all data share a
        common ctype).  If multiple data ctypes are found or type is
        ``None``, then :py:class:`IndexedComponent` will be used.

    Examples
    --------

    .. doctest::

        >>> from pyomo.environ import *
        >>> m = ConcreteModel()
        >>> @m.Block([1,2],[3,4])
        ... def b(b,i,j):
        ...     b.x = Var(bounds=(i,j))
        ...
        >>> m.r1 = Reference(m.b[:,:].x)
        >>> m.r1.pprint()
        r1 : Size=4, Index={1, 2}*{3, 4}, ReferenceTo=b[:, :].x
            Key    : Lower : Value : Upper : Fixed : Stale : Domain
            (1, 3) :     1 :  None :     3 : False :  True :  Reals
            (1, 4) :     1 :  None :     4 : False :  True :  Reals
            (2, 3) :     2 :  None :     3 : False :  True :  Reals
            (2, 4) :     2 :  None :     4 : False :  True :  Reals

    Reference components may also refer to subsets of the original data:

    .. doctest::

        >>> m.r2 = Reference(m.b[:,3].x)
        >>> m.r2.pprint()
        r2 : Size=2, Index={1, 2}, ReferenceTo=b[:, 3].x
            Key : Lower : Value : Upper : Fixed : Stale : Domain
              1 :     1 :  None :     3 : False :  True :  Reals
              2 :     2 :  None :     3 : False :  True :  Reals

    Reference components may have wildcards at multiple levels of the
    model hierarchy:

    .. doctest::

        >>> m = ConcreteModel()
        >>> @m.Block([1,2])
        ... def b(b,i):
        ...     b.x = Var([3,4], bounds=(i,None))
        ...
        >>> m.r3 = Reference(m.b[:].x[:])
        >>> m.r3.pprint()
        r3 : Size=4, Index=ReferenceSet(b[:].x[:]), ReferenceTo=b[:].x[:]
            Key    : Lower : Value : Upper : Fixed : Stale : Domain
            (1, 3) :     1 :  None :  None : False :  True :  Reals
            (1, 4) :     1 :  None :  None : False :  True :  Reals
            (2, 3) :     2 :  None :  None : False :  True :  Reals
            (2, 4) :     2 :  None :  None : False :  True :  Reals

    The resulting reference component may be used just like any other
    component.  Changes to the stored data will be reflected in the
    original objects:

    .. doctest::

        >>> m.r3[1,4] = 10
        >>> m.b[1].x.pprint()
        x : Size=2, Index={3, 4}
            Key : Lower : Value : Upper : Fixed : Stale : Domain
              3 :     1 :  None :  None : False :  True :  Reals
              4 :     1 :    10 :  None : False : False :  Reals

    """
    referent = reference
    #
    # Before constructing the reference object that we will return,
    # we need to know its index set, its ctype, and its _data
    # dict. The following if statement sets the _data dict
    # for all possible input types and sets up data structures
    # necessary to determine the index set and ctype.
    #
    if isinstance(reference, IndexedComponent_slice):
        slice_idx = []
        index = None
        _data = _ReferenceDict(reference)
        _iter = iter(reference)
    elif isinstance(reference, Component):
        slice_idx = None
        if reference.is_indexed():
            index = reference.index_set()
        else:
            index = UnindexedComponent_ReferenceSet
        if ctype is NOTSET:
            ctype = reference.ctype
        reference = reference[...]
        _data = _ReferenceDict(reference)
        # index and ctype are now set; no need to iterate over the slice
        _iter = ()
    elif isinstance(reference, ComponentData):
        slice_idx = None
        index = UnindexedComponent_ReferenceSet
        if ctype is NOTSET:
            ctype = reference.ctype
        reference = IndexedComponent_slice(reference.parent_component())[
            reference.index()
        ]
        _data = _ReferenceDict(reference)
        # index and ctype are now set; no need to iterate over the slice
        _iter = ()
    elif isinstance(reference, Mapping):
        slice_idx = None
        _data = _ReferenceDict_mapping(dict(reference))
        _iter = _data.values()
        index = SetOf(_data)
    elif isinstance(reference, Sequence):
        slice_idx = None
        _data = _ReferenceDict_mapping(dict(enumerate(reference)))
        _iter = _data.values()
        index = OrderedSetOf(_data)
    else:
        raise TypeError(
            "First argument to Reference constructors must be a "
            "component, component slice, Sequence, or Mapping (received %s)"
            % (type(reference).__name__,)
        )

    if ctype is NOTSET:
        ctypes = set()
    else:
        if slice_idx is None:
            # A slice was not provided. We know the ctype and that there
            # cannot be common subsets (because a slice was not provided).
            # We don't need to iterate over the data objects at all.
            # Note that this is redundant for Component and ComponentData
            # inputs, as _iter was already empty.
            _iter = ()

    for obj in _iter:
        #
        # We were provided a collection of ComponentData, either via
        # a slice, sequence, or mapping. Now we iterate over these
        # objects to attempt to infer both the ctype and, if a slice
        # was provided, the index_set (determined by slice_index).
        #
        if ctype is NOTSET:
            ctypes.add(obj.ctype)
            if not isinstance(obj, ComponentData):
                # This object is not a ComponentData (likely it is a pure
                # IndexedComponent container).  As the Reference will treat
                # it as if it *were* a ComponentData, we will skip ctype
                # identification and return a base IndexedComponent, thereby
                # preventing strange exceptions in the writers and with
                # things like pprint().  Of course, all of this logic is
                # skipped if the User knows better and forced a ctype on us.
                ctypes.add(0)
        # Note that we want to walk the entire slice, unless we can
        # prove that BOTH there aren't common indexing sets (i.e., index
        # is None) AND there is more than one ctype.
        if slice_idx is not None:
            # As long as we haven't ruled out the possibility of common
            # wildcard sets, then we will use _identify_wildcard_sets to
            # identify the wildcards for this obj and check compatibility
            # of the wildcards with any previously-identified wildcards.
            slice_idx = _identify_wildcard_sets(_iter._iter_stack, slice_idx)
        elif ctype is not NOTSET or len(ctypes) > 1:
            break

    if index is None:
        #
        # index is None, i.e. a slice was provided. If a slice index
        # has been identified (by all slice members having the same
        # "wildcard sets"), use this to construct an indexing set.
        # Otherwise, use a _ReferenceSet.
        #
        if not slice_idx:
            index = SetOf(_ReferenceSet(reference))
        else:
            wildcards = sum(
                (sorted(lvl.items()) for lvl in slice_idx if lvl is not None), []
            )
            # Wildcards is a list of (coordinate, set) tuples.  Coordinate
            # is that within the subsets list, and set is a wildcard set.
            index = wildcards[0][1]
            # index is the first wildcard set.
            if not isinstance(index, SetData):
                index = SetOf(index)
            for lvl, idx in wildcards[1:]:
                if not isinstance(idx, SetData):
                    idx = SetOf(idx)
                index = index * idx
            # index is now either a single Set, or a SetProduct of the
            # wildcard sets.
    if ctype is NOTSET:
        if len(ctypes) == 1:
            # If ctype is not set and only one ctype was identified above,
            # use this ctype.
            ctype = ctypes.pop()
        else:
            ctype = IndexedComponent
    elif ctype is None:
        ctype = IndexedComponent

    obj = ctype(index, ctype=ctype)
    obj._constructed = True
    obj._data = _data
    obj.referent = referent
    return obj



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import io
from typing import List, Optional

from pyomo.common.collections import ComponentMap
from pyomo.common.dependencies import attempt_import
from pyomo.common.errors import ApplicationError
from pyomo.common.flags import NOTSET
from pyomo.common.tee import TeeStream, capture_output
from pyomo.core.kernel.objective import minimize, maximize
from pyomo.core.base.var import VarData
from pyomo.core.base.constraint import ConstraintData
from pyomo.core.base.sos import SOSConstraintData
from pyomo.core.base.param import ParamData
from pyomo.core.expr.numvalue import value, is_constant
from pyomo.repn import generate_standard_repn
from pyomo.core.expr.numeric_expr import NPV_MaxExpression, NPV_MinExpression
from pyomo.common.dependencies import numpy as np
from pyomo.core.staleflag import StaleFlagManager

from pyomo.contrib.solver.common.base import PersistentSolverBase, Availability
from pyomo.contrib.solver.common.results import (
    Results,
    TerminationCondition,
    SolutionStatus,
)
from pyomo.contrib.solver.common.config import PersistentBranchAndBoundConfig
from pyomo.contrib.solver.common.persistent import (
    PersistentSolverUtils,
    PersistentSolverMixin,
)
from pyomo.contrib.solver.common.solution_loader import PersistentSolutionLoader
from pyomo.contrib.solver.common.util import (
    NoFeasibleSolutionError,
    NoOptimalSolutionError,
    NoDualsError,
    NoReducedCostsError,
    NoSolutionError,
    IncompatibleModelError,
)

logger = logging.getLogger(__name__)

highspy, highspy_available = attempt_import('highspy')


class _MutableVarBounds:
    def __init__(self, lower_expr, upper_expr, pyomo_var_id, var_map, highs):
        self.pyomo_var_id = pyomo_var_id
        self.lower_expr = lower_expr
        self.upper_expr = upper_expr
        self.var_map = var_map
        self.highs = highs

    def update(self):
        col_ndx = self.var_map[self.pyomo_var_id]
        lb = value(self.lower_expr)
        ub = value(self.upper_expr)
        self.highs.changeColBounds(col_ndx, lb, ub)


class _MutableLinearCoefficient:
    def __init__(self, pyomo_con, pyomo_var_id, con_map, var_map, expr, highs):
        self.expr = expr
        self.highs = highs
        self.pyomo_var_id = pyomo_var_id
        self.pyomo_con = pyomo_con
        self.con_map = con_map
        self.var_map = var_map

    def update(self):
        row_ndx = self.con_map[self.pyomo_con]
        col_ndx = self.var_map[self.pyomo_var_id]
        self.highs.changeCoeff(row_ndx, col_ndx, value(self.expr))


class _MutableObjectiveCoefficient:
    def __init__(self, pyomo_var_id, var_map, expr, highs):
        self.expr = expr
        self.highs = highs
        self.pyomo_var_id = pyomo_var_id
        self.var_map = var_map

    def update(self):
        col_ndx = self.var_map[self.pyomo_var_id]
        self.highs.changeColCost(col_ndx, value(self.expr))


class _MutableQuadraticCoefficient:
    def __init__(self, expr, v1_id, v2_id):
        self.expr = expr
        self.v1_id = v1_id
        self.v2_id = v2_id


class _MutableObjective:
    def __init__(
        self,
        highs,
        constant,
        linear_coefs,
        quadratic_coefs,
        pyomo_var_to_solver_var_map,
    ):
        self.highs = highs
        self.constant = constant
        self.linear_coefs = linear_coefs
        self.quadratic_coefs = quadratic_coefs
        self._pyomo_var_to_solver_var_map = pyomo_var_to_solver_var_map
        # Store the quadratic coefficients in dictionary format
        self._initialize_quad_coef_dicts()
        # Flag to force first update of quadratic coefficients
        self._first_update = True

    def _initialize_quad_coef_dicts(self):
        self.quad_coef_dict = {}
        for coef in self.quadratic_coefs:
            self.quad_coef_dict[(coef.v1_id, coef.v2_id)] = value(coef.expr)
        self.previous_quad_coef_dict = self.quad_coef_dict.copy()

    def update(self):
        """
        Update the quadratic objective expression.
        """
        needs_quadratic_update = self._first_update

        self.constant.update()
        for coef in self.linear_coefs:
            coef.update()

        for coef in self.quadratic_coefs:
            current_val = value(coef.expr)
            previous_val = self.previous_quad_coef_dict.get((coef.v1_id, coef.v2_id))
            if previous_val is not None and current_val != previous_val:
                needs_quadratic_update = True
                self.quad_coef_dict[(coef.v1_id, coef.v2_id)] = current_val
                self.previous_quad_coef_dict[(coef.v1_id, coef.v2_id)] = current_val

        # If anything changed, rebuild and pass the Hessian
        if needs_quadratic_update:
            self._build_and_pass_hessian()
            self._first_update = False

    def _build_and_pass_hessian(self):
        """Build and pass the Hessian to HiGHS in CSC format"""
        if not self.quad_coef_dict:
            return

        dim = self.highs.getNumCol()

        # Build CSC format for the lower triangular part
        hessian_value = []
        hessian_index = []
        hessian_start = [0] * dim

        quad_coef_idx_dict = {}
        for (v1_id, v2_id), coef in self.quad_coef_dict.items():
            v1_ndx = self._pyomo_var_to_solver_var_map[v1_id]
            v2_ndx = self._pyomo_var_to_solver_var_map[v2_id]
            # Ensure we're storing the lower triangular part
            row = max(v1_ndx, v2_ndx)
            col = min(v1_ndx, v2_ndx)
            # Adjust the diagonal to match Highs' expected format
            if v1_ndx == v2_ndx:
                coef *= 2.0
            quad_coef_idx_dict[(row, col)] = coef

        sorted_entries = sorted(
            quad_coef_idx_dict.items(), key=lambda x: (x[0][1], x[0][0])
        )

        last_col = -1
        for (row, col), val in sorted_entries:
            while col > last_col:
                last_col += 1
                if last_col < dim:
                    hessian_start[last_col] = len(hessian_value)

            # Add the entry
            hessian_index.append(row)
            hessian_value.append(val)

        while last_col < dim - 1:
            last_col += 1
            hessian_start[last_col] = len(hessian_value)

        nnz = len(hessian_value)
        status = self.highs.passHessian(
            dim,
            nnz,
            highspy.HessianFormat.kTriangular,
            np.array(hessian_start, dtype=np.int32),
            np.array(hessian_index, dtype=np.int32),
            np.array(hessian_value, dtype=np.double),
        )

        if status != highspy.HighsStatus.kOk:
            logger.warning(
                f"HiGHS returned non-OK status when passing Hessian: {status}"
            )


class _MutableObjectiveOffset:
    def __init__(self, expr, highs):
        self.expr = expr
        self.highs = highs

    def update(self):
        self.highs.changeObjectiveOffset(value(self.expr))


class _MutableConstraintBounds:
    def __init__(self, lower_expr, upper_expr, pyomo_con, con_map, highs):
        self.lower_expr = lower_expr
        self.upper_expr = upper_expr
        self.con = pyomo_con
        self.con_map = con_map
        self.highs = highs

    def update(self):
        row_ndx = self.con_map[self.con]
        lb = value(self.lower_expr)
        ub = value(self.upper_expr)
        self.highs.changeRowBounds(row_ndx, lb, ub)


class Highs(PersistentSolverMixin, PersistentSolverUtils, PersistentSolverBase):
    """
    Interface to HiGHS
    """

    CONFIG = PersistentBranchAndBoundConfig()

    _available = None

    def __init__(self, **kwds):
        treat_fixed_vars_as_params = kwds.pop('treat_fixed_vars_as_params', True)
        PersistentSolverBase.__init__(self, **kwds)
        PersistentSolverUtils.__init__(
            self, treat_fixed_vars_as_params=treat_fixed_vars_as_params
        )
        self._solver_model = None
        self._pyomo_var_to_solver_var_map = {}
        self._pyomo_con_to_solver_con_map = {}
        self._solver_con_to_pyomo_con_map = {}
        self._mutable_helpers = {}
        self._mutable_bounds = {}
        self._last_results_object: Optional[Results] = None
        self._sol = None

    def available(self):
        if highspy_available:
            return Availability.FullLicense
        return Availability.NotFound

    def version(self):
        try:
            version = (
                highspy.HIGHS_VERSION_MAJOR,
                highspy.HIGHS_VERSION_MINOR,
                highspy.HIGHS_VERSION_PATCH,
            )
        except AttributeError:
            # Older versions of Highs do not have the above attributes
            # and the solver version can only be obtained by making
            # an instance of the solver class.
            tmp = highspy.Highs()
            version = (tmp.versionMajor(), tmp.versionMinor(), tmp.versionPatch())

        return version

    def _solve(self):
        config = self._active_config
        timer = config.timer
        options = config.solver_options
        ostreams = [io.StringIO()] + config.tee

        with capture_output(output=TeeStream(*ostreams), capture_fd=True):
            self._solver_model.setOptionValue('log_to_console', True)

            if config.threads is not None:
                self._solver_model.setOptionValue('threads', config.threads)
            if config.time_limit is not None:
                self._solver_model.setOptionValue('time_limit', config.time_limit)
            if config.rel_gap is not None:
                self._solver_model.setOptionValue('mip_rel_gap', config.rel_gap)
            if config.abs_gap is not None:
                self._solver_model.setOptionValue('mip_abs_gap', config.abs_gap)

            for key, option in options.items():
                self._solver_model.setOptionValue(key, option)
            timer.start('optimize')
            if self.version()[:2] >= (1, 8):
                self._solver_model.HandleKeyboardInterrupt = True
            self._solver_model.run()
            timer.stop('optimize')

        return self._postsolve(ostreams[0])

    def _process_domain_and_bounds(self, var_id):
        _v, _lb, _ub, _fixed, _domain_interval, _value = self._vars[var_id]
        lb, ub, step = _domain_interval
        if lb is None:
            lb = -highspy.kHighsInf
        if ub is None:
            ub = highspy.kHighsInf
        if step == 0:
            vtype = highspy.HighsVarType.kContinuous
        elif step == 1:
            vtype = highspy.HighsVarType.kInteger
        else:
            raise ValueError(
                f'Unrecognized domain step: {step} (should be either 0 or 1)'
            )
        if _fixed:
            lb = _value
            ub = _value
        else:
            if _lb is not None or _ub is not None:
                if not is_constant(_lb) or not is_constant(_ub):
                    if _lb is None:
                        tmp_lb = -highspy.kHighsInf
                    else:
                        tmp_lb = _lb
                    if _ub is None:
                        tmp_ub = highspy.kHighsInf
                    else:
                        tmp_ub = _ub
                    mutable_bound = _MutableVarBounds(
                        lower_expr=NPV_MaxExpression((tmp_lb, lb)),
                        upper_expr=NPV_MinExpression((tmp_ub, ub)),
                        pyomo_var_id=var_id,
                        var_map=self._pyomo_var_to_solver_var_map,
                        highs=self._solver_model,
                    )
                    self._mutable_bounds[var_id] = (_v, mutable_bound)
            if _lb is not None:
                lb = max(value(_lb), lb)
            if _ub is not None:
                ub = min(value(_ub), ub)

        return lb, ub, vtype

    def _add_variables(self, variables: List[VarData]):
        self._sol = None
        if self._last_results_object is not None:
            self._last_results_object.solution_loader.invalidate()
        lbs = []
        ubs = []
        indices = []
        vtypes = []

        current_num_vars = len(self._pyomo_var_to_solver_var_map)
        for v in variables:
            v_id = id(v)
            lb, ub, vtype = self._process_domain_and_bounds(v_id)
            lbs.append(lb)
            ubs.append(ub)
            vtypes.append(vtype)
            indices.append(current_num_vars)
            self._pyomo_var_to_solver_var_map[v_id] = current_num_vars
            current_num_vars += 1

        self._solver_model.addVars(
            len(lbs), np.array(lbs, dtype=np.double), np.array(ubs, dtype=np.double)
        )
        self._solver_model.changeColsIntegrality(
            len(vtypes), np.array(indices), np.array(vtypes)
        )

    def _add_parameters(self, params: List[ParamData]):
        pass

    def _reinit(self):
        saved_config = self.config
        saved_active_config = self._active_config
        self.__init__(treat_fixed_vars_as_params=self._treat_fixed_vars_as_params)
        self.config = saved_config
        self._active_config = saved_active_config

    def set_instance(self, model):
        config = self._active_config
        ostreams = config.tee

        if self._last_results_object is not None:
            self._last_results_object.solution_loader.invalidate()
        if not self.available():
            c = self.__class__
            raise ApplicationError(
                f'Solver {c.__module__}.{c.__qualname__} is not available '
                f'({self.available()}).'
            )

        with capture_output(TeeStream(*ostreams), capture_fd=True):
            self._reinit()
            self._model = model

            self._solver_model = highspy.Highs()
            self.add_block(model)
            if self._objective is None:
                self.set_objective(None)

    def _add_constraints(self, cons: List[ConstraintData]):
        self._sol = None
        if self._last_results_object is not None:
            self._last_results_object.solution_loader.invalidate()
        current_num_cons = len(self._pyomo_con_to_solver_con_map)
        lbs = []
        ubs = []
        starts = []
        var_indices = []
        coef_values = []

        for con in cons:
            repn = generate_standard_repn(
                con.body, quadratic=False, compute_values=False
            )
            if repn.nonlinear_expr is not None:
                raise IncompatibleModelError(
                    f'Highs interface does not support expressions of degree {repn.polynomial_degree()}'
                )

            starts.append(len(coef_values))
            for ndx, coef in enumerate(repn.linear_coefs):
                v = repn.linear_vars[ndx]
                v_id = id(v)
                coef_val = value(coef)
                if not is_constant(coef):
                    mutable_linear_coefficient = _MutableLinearCoefficient(
                        pyomo_con=con,
                        pyomo_var_id=v_id,
                        con_map=self._pyomo_con_to_solver_con_map,
                        var_map=self._pyomo_var_to_solver_var_map,
                        expr=coef,
                        highs=self._solver_model,
                    )
                    if con not in self._mutable_helpers:
                        self._mutable_helpers[con] = []
                    self._mutable_helpers[con].append(mutable_linear_coefficient)
                    if coef_val == 0:
                        continue
                var_indices.append(self._pyomo_var_to_solver_var_map[v_id])
                coef_values.append(coef_val)

            if con.has_lb():
                lb = con.lower - repn.constant
            else:
                lb = -highspy.kHighsInf
            if con.has_ub():
                ub = con.upper - repn.constant
            else:
                ub = highspy.kHighsInf

            if not is_constant(lb) or not is_constant(ub):
                mutable_con_bounds = _MutableConstraintBounds(
                    lower_expr=lb,
                    upper_expr=ub,
                    pyomo_con=con,
                    con_map=self._pyomo_con_to_solver_con_map,
                    highs=self._solver_model,
                )
                if con not in self._mutable_helpers:
                    self._mutable_helpers[con] = [mutable_con_bounds]
                else:
                    self._mutable_helpers[con].append(mutable_con_bounds)

            lbs.append(value(lb))
            ubs.append(value(ub))
            self._pyomo_con_to_solver_con_map[con] = current_num_cons
            self._solver_con_to_pyomo_con_map[current_num_cons] = con
            current_num_cons += 1

        self._solver_model.addRows(
            len(lbs),
            np.array(lbs, dtype=np.double),
            np.array(ubs, dtype=np.double),
            len(coef_values),
            np.array(starts),
            np.array(var_indices),
            np.array(coef_values, dtype=np.double),
        )

    def _add_sos_constraints(self, cons: List[SOSConstraintData]):
        if cons:
            raise NotImplementedError(
                'Highs interface does not support SOS constraints'
            )

    def _remove_constraints(self, cons: List[ConstraintData]):
        self._sol = None
        if self._last_results_object is not None:
            self._last_results_object.solution_loader.invalidate()
        indices_to_remove = []
        for con in cons:
            con_ndx = self._pyomo_con_to_solver_con_map.pop(con)
            del self._solver_con_to_pyomo_con_map[con_ndx]
            indices_to_remove.append(con_ndx)
            self._mutable_helpers.pop(con, None)
        self._solver_model.deleteRows(
            len(indices_to_remove), np.array(list(sorted(indices_to_remove)))
        )
        con_ndx = 0
        new_con_map = {}
        for c in self._pyomo_con_to_solver_con_map:
            new_con_map[c] = con_ndx
            con_ndx += 1
        self._pyomo_con_to_solver_con_map.clear()
        self._pyomo_con_to_solver_con_map.update(new_con_map)
        self._solver_con_to_pyomo_con_map.clear()
        self._solver_con_to_pyomo_con_map.update(
            {v: k for k, v in self._pyomo_con_to_solver_con_map.items()}
        )

    def _remove_sos_constraints(self, cons: List[SOSConstraintData]):
        if cons:
            raise NotImplementedError(
                'Highs interface does not support SOS constraints'
            )

    def _remove_variables(self, variables: List[VarData]):
        self._sol = None
        if self._last_results_object is not None:
            self._last_results_object.solution_loader.invalidate()
        indices_to_remove = []
        for v in variables:
            v_id = id(v)
            v_ndx = self._pyomo_var_to_solver_var_map.pop(v_id)
            indices_to_remove.append(v_ndx)
            self._mutable_bounds.pop(v_id, None)
        indices_to_remove.sort()
        self._solver_model.deleteVars(
            len(indices_to_remove), np.array(list(sorted(indices_to_remove)))
        )
        v_ndx = 0
        new_var_map = {}
        for v_id in self._pyomo_var_to_solver_var_map:
            new_var_map[v_id] = v_ndx
            v_ndx += 1
        self._pyomo_var_to_solver_var_map.clear()
        self._pyomo_var_to_solver_var_map.update(new_var_map)

    def _remove_parameters(self, params: List[ParamData]):
        pass

    def _update_variables(self, variables: List[VarData]):
        self._sol = None
        if self._last_results_object is not None:
            self._last_results_object.solution_loader.invalidate()
        indices = []
        lbs = []
        ubs = []
        vtypes = []

        for v in variables:
            v_id = id(v)
            self._mutable_bounds.pop(v_id, None)
            v_ndx = self._pyomo_var_to_solver_var_map[v_id]
            lb, ub, vtype = self._process_domain_and_bounds(v_id)
            lbs.append(lb)
            ubs.append(ub)
            vtypes.append(vtype)
            indices.append(v_ndx)

        self._solver_model.changeColsBounds(
            len(indices),
            np.array(indices),
            np.array(lbs, dtype=np.double),
            np.array(ubs, dtype=np.double),
        )
        self._solver_model.changeColsIntegrality(
            len(indices), np.array(indices), np.array(vtypes)
        )

    def update_parameters(self):
        self._sol = None
        if self._last_results_object is not None:
            self._last_results_object.solution_loader.invalidate()

        for con, helpers in self._mutable_helpers.items():
            for helper in helpers:
                helper.update()
        for k, (v, helper) in self._mutable_bounds.items():
            helper.update()

        self._mutable_objective.update()

    def _set_objective(self, obj):
        self._sol = None
        if self._last_results_object is not None:
            self._last_results_object.solution_loader.invalidate()
        n = len(self._pyomo_var_to_solver_var_map)
        indices = np.arange(n)
        costs = np.zeros(n, dtype=np.double)

        # Initialize empty lists for all coefficient types
        mutable_linear_coefficients = []
        mutable_quadratic_coefficients = []

        if obj is None:
            sense = highspy.ObjSense.kMinimize
            mutable_constant = _MutableObjectiveOffset(expr=0, highs=self._solver_model)
        else:
            if obj.sense == minimize:
                sense = highspy.ObjSense.kMinimize
            elif obj.sense == maximize:
                sense = highspy.ObjSense.kMaximize
            else:
                raise ValueError(f'Objective sense is not recognized: {obj.sense}')

            repn = generate_standard_repn(
                obj.expr, quadratic=True, compute_values=False
            )
            if repn.nonlinear_expr is not None or repn.polynomial_degree() > 2:
                raise IncompatibleModelError(
                    f'Highs interface does not support expressions of degree {repn.polynomial_degree()}'
                )

            for coef, v in zip(repn.linear_coefs, repn.linear_vars):
                v_id = id(v)
                v_ndx = self._pyomo_var_to_solver_var_map[v_id]
                costs[v_ndx] = value(coef)
                if not is_constant(coef):
                    mutable_objective_coef = _MutableObjectiveCoefficient(
                        pyomo_var_id=v_id,
                        var_map=self._pyomo_var_to_solver_var_map,
                        expr=coef,
                        highs=self._solver_model,
                    )
                    mutable_linear_coefficients.append(mutable_objective_coef)

            mutable_constant = _MutableObjectiveOffset(
                expr=repn.constant, highs=self._solver_model
            )

            if repn.quadratic_vars and len(repn.quadratic_vars) > 0:
                for ndx, (v1, v2) in enumerate(repn.quadratic_vars):
                    coef = repn.quadratic_coefs[ndx]

                    mutable_quadratic_coefficients.append(
                        _MutableQuadraticCoefficient(
                            expr=coef, v1_id=id(v1), v2_id=id(v2)
                        )
                    )

        self._solver_model.changeObjectiveSense(sense)
        self._solver_model.changeColsCost(n, indices, costs)
        self._mutable_objective = _MutableObjective(
            self._solver_model,
            mutable_constant,
            mutable_linear_coefficients,
            mutable_quadratic_coefficients,
            self._pyomo_var_to_solver_var_map,
        )
        self._mutable_objective.update()

    def _postsolve(self, stream: io.StringIO):
        config = self._active_config
        timer = config.timer
        timer.start('load solution')

        highs = self._solver_model
        status = highs.getModelStatus()

        results = Results()
        results.solution_loader = PersistentSolutionLoader(self)
        results.solver_name = self.name
        results.solver_version = self.version()
        results.solver_config = config
        results.solver_log = stream.getvalue()
        results.timing_info.highs_time = highs.getRunTime()

        self._sol = highs.getSolution()
        has_feasible_solution = self._sol.value_valid
        if status == highspy.HighsModelStatus.kOptimal:
            results.solution_status = SolutionStatus.optimal
        elif has_feasible_solution:
            results.solution_status = SolutionStatus.feasible
        else:
            results.solution_status = SolutionStatus.noSolution

        if status == highspy.HighsModelStatus.kNotset:
            results.termination_condition = TerminationCondition.unknown
        elif status == highspy.HighsModelStatus.kLoadError:
            results.termination_condition = TerminationCondition.error
        elif status == highspy.HighsModelStatus.kModelError:
            results.termination_condition = TerminationCondition.error
        elif status == highspy.HighsModelStatus.kPresolveError:
            results.termination_condition = TerminationCondition.error
        elif status == highspy.HighsModelStatus.kSolveError:
            results.termination_condition = TerminationCondition.error
        elif status == highspy.HighsModelStatus.kPostsolveError:
            results.termination_condition = TerminationCondition.error
        elif status == highspy.HighsModelStatus.kModelEmpty:
            results.termination_condition = TerminationCondition.unknown
        elif status == highspy.HighsModelStatus.kOptimal:
            results.termination_condition = (
                TerminationCondition.convergenceCriteriaSatisfied
            )
        elif status == highspy.HighsModelStatus.kInfeasible:
            results.termination_condition = TerminationCondition.provenInfeasible
        elif status == highspy.HighsModelStatus.kUnboundedOrInfeasible:
            results.termination_condition = TerminationCondition.infeasibleOrUnbounded
        elif status == highspy.HighsModelStatus.kUnbounded:
            results.termination_condition = TerminationCondition.unbounded
        elif status == highspy.HighsModelStatus.kObjectiveBound:
            results.termination_condition = TerminationCondition.objectiveLimit
        elif status == highspy.HighsModelStatus.kObjectiveTarget:
            results.termination_condition = TerminationCondition.objectiveLimit
        elif status == highspy.HighsModelStatus.kTimeLimit:
            results.termination_condition = TerminationCondition.maxTimeLimit
        elif status == highspy.HighsModelStatus.kIterationLimit:
            results.termination_condition = TerminationCondition.iterationLimit
        elif status == getattr(highspy.HighsModelStatus, "kSolutionLimit", NOTSET):
            # kSolutionLimit was introduced in HiGHS v1.5.3 for MIP-related limits
            results.termination_condition = TerminationCondition.iterationLimit
        elif status == highspy.HighsModelStatus.kUnknown:
            results.termination_condition = TerminationCondition.unknown
        else:
            logger.warning(f'Received unhandled {status=} from solver HiGHS.')
            results.termination_condition = TerminationCondition.unknown

        if (
            results.termination_condition
            != TerminationCondition.convergenceCriteriaSatisfied
            and config.raise_exception_on_nonoptimal_result
        ):
            raise NoOptimalSolutionError()

        results.incumbent_objective = None
        results.objective_bound = None
        info = highs.getInfo()
        if self._objective is not None:
            if has_feasible_solution:
                results.incumbent_objective = info.objective_function_value
            if info.mip_node_count == -1:
                if has_feasible_solution:
                    results.objective_bound = info.objective_function_value
                else:
                    results.objective_bound = None
            else:
                results.objective_bound = info.mip_dual_bound

            if info.valid:
                results.extra_info.simplex_iteration_count = (
                    info.simplex_iteration_count
                )
                results.extra_info.ipm_iteration_count = info.ipm_iteration_count
                results.extra_info.mip_node_count = info.mip_node_count
                results.extra_info.pdlp_iteration_count = info.pdlp_iteration_count
                results.extra_info.qp_iteration_count = info.qp_iteration_count

        if config.load_solutions:
            if has_feasible_solution:
                self._load_vars()
            else:
                raise NoFeasibleSolutionError()
        timer.stop('load solution')

        return results

    def _load_vars(self, vars_to_load=None):
        for v, val in self._get_primals(vars_to_load=vars_to_load).items():
            v.set_value(val, skip_validation=True)
        StaleFlagManager.mark_all_as_stale(delayed=True)

    def _get_primals(self, vars_to_load=None):
        if self._sol is None or not self._sol.value_valid:
            raise NoSolutionError()

        res = ComponentMap()
        if vars_to_load is None:
            var_ids_to_load = []
            for v, ref_info in self._referenced_variables.items():
                using_cons, using_sos, using_obj = ref_info
                if using_cons or using_sos or (using_obj is not None):
                    var_ids_to_load.append(v)
        else:
            var_ids_to_load = [id(v) for v in vars_to_load]

        var_vals = self._sol.col_value

        for v_id in var_ids_to_load:
            v = self._vars[v_id][0]
            v_ndx = self._pyomo_var_to_solver_var_map[v_id]
            res[v] = var_vals[v_ndx]

        return res

    def _get_reduced_costs(self, vars_to_load=None):
        if self._sol is None or not self._sol.dual_valid:
            raise NoReducedCostsError()
        res = ComponentMap()
        if vars_to_load is None:
            var_ids_to_load = list(self._vars.keys())
        else:
            var_ids_to_load = [id(v) for v in vars_to_load]

        var_vals = self._sol.col_dual

        for v_id in var_ids_to_load:
            v = self._vars[v_id][0]
            v_ndx = self._pyomo_var_to_solver_var_map[v_id]
            res[v] = var_vals[v_ndx]

        return res

    def _get_duals(self, cons_to_load=None):
        if self._sol is None or not self._sol.dual_valid:
            raise NoDualsError()

        res = {}
        if cons_to_load is None:
            cons_to_load = list(self._pyomo_con_to_solver_con_map.keys())

        duals = self._sol.row_dual

        for c in cons_to_load:
            c_ndx = self._pyomo_con_to_solver_con_map[c]
            res[c] = duals[c_ndx]

        return res



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os

import subprocess
import pyomo.common
from pyomo.common.errors import ApplicationError
from pyomo.common.tempfiles import TempfileManager
from pyomo.opt.base import ProblemFormat, ConverterError
from pyomo.opt.base.convert import ProblemConverterFactory


@ProblemConverterFactory.register('glpsol')
class GlpsolMIPConverter:
    def can_convert(self, from_type, to_type):
        """Returns true if this object supports the specified conversion"""
        #
        # Test if the glpsol executable is available
        #
        if not pyomo.common.Executable("glpsol"):
            return False
        #
        # Return True for specific from/to pairs
        #
        if from_type == ProblemFormat.mod and to_type == ProblemFormat.cpxlp:
            return True
        if from_type == ProblemFormat.mod and to_type == ProblemFormat.mps:
            if not pyomo.common.Executable("ampl"):
                #
                # Only convert mod->mps with ampl
                #
                return True
        return False

    def apply(self, *args, **kwargs):
        """Convert an instance of one type into another"""
        if not isinstance(args[2], str):
            raise ConverterError("Can only apply glpsol to convert file data")
        _exec = pyomo.common.Executable("glpsol")
        if not _exec:
            raise ConverterError("The 'glpsol' executable cannot be found")
        cmd = [_exec.path(), "--math"]
        #
        # MPS->LP conversion is ignored in coverage because it's not being
        #   used; instead, we're using pico_convert for this conversion
        #
        modfile = ''
        if args[1] == ProblemFormat.mps:  # pragma:nocover
            ofile = TempfileManager.create_tempfile(suffix='.glpsol.mps')
            cmd.extend(
                [
                    "--check",
                    "--name",
                    "MPS model derived from " + os.path.basename(args[2]),
                    "--wfreemps",
                    ofile,
                ]
            )
        elif args[1] == ProblemFormat.cpxlp:
            ofile = TempfileManager.create_tempfile(suffix='.glpsol.lp')
            cmd.extend(
                [
                    "--check",
                    "--name",
                    "MPS model derived from " + os.path.basename(args[2]),
                    "--wcpxlp",
                    ofile,
                ]
            )
        if len(args[2:]) == 1:
            cmd.append(args[2])
        else:
            #
            # Create a temporary model file, since GLPSOL can only
            # handle one input file
            #
            modfile = TempfileManager.create_tempfile(suffix='.glpsol.mod')
            OUTPUT = open(modfile, "w")
            flag = False
            #
            # Read the model file
            #
            INPUT = open(args[2])
            for line in INPUT:
                line = line.strip()
                if line == "data;":
                    raise ConverterError(
                        "Problem composing mathprog model and data files - mathprog file already has data in it!"
                    )
                if line != "end;":
                    OUTPUT.write(line + '\n')
            INPUT.close()
            OUTPUT.write("data;\n")
            #
            # Read the data files
            #
            for file in args[3:]:
                INPUT = open(file)
                for line in INPUT:
                    line = line.strip()
                    if line != "end;" and line != "data;":
                        OUTPUT.write(line + '\n')
                INPUT.close()
                OUTPUT.write("end;\n")
            OUTPUT.close()
            cmd.append(modfile)
        subprocess.run(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        if not os.path.exists(ofile):  # pragma:nocover
            raise ApplicationError("Problem launching 'glpsol' to create " + ofile)
        if os.path.exists(modfile):
            os.remove(modfile)
        return (ofile,), None  # empty variable map



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.cp.interval_var import (
    IntervalVar,
    IntervalVarData,
    IntervalVarStartTime,
    IntervalVarEndTime,
)
from pyomo.core.expr.base import ExpressionBase
from pyomo.core.expr.logical_expr import BooleanExpression
from pyomo.core.expr.numeric_expr import SumExpression


def _sum_two_units(_self, _other):
    return CumulativeFunction([_self, _other])


def _sum_cumul_and_unit(_cumul, _unit):
    if _cumul.nargs() == len(_cumul._args_):
        # we can just append to the cumul list
        _cumul._args_.append(_unit)
        return CumulativeFunction(_cumul._args_, nargs=len(_cumul._args_))
    else:
        return CumulativeFunction(_cumul.args + [_unit])


def _sum_unit_and_cumul(_unit, _cumul):
    # Nothing to be done: we need to make a new one because we can't prepend to
    # the list of args.
    return CumulativeFunction([_unit] + _cumul.args)


def _sum_cumuls(_self, _other):
    if _self.nargs() == len(_self._args_):
        _self._args_.extend(_other.args)
        return CumulativeFunction(_self._args_, nargs=len(_self._args_))
    else:
        # we have to clone the list of _args_
        return CumulativeFunction(_self.args + _other.args)


def _subtract_two_units(_self, _other):
    return CumulativeFunction([_self, NegatedStepFunction((_other,))])


def _subtract_cumul_and_unit(_cumul, _unit):
    if _cumul.nargs() == len(_cumul._args_):
        # we can just append to the cumul list
        _cumul._args_.append(NegatedStepFunction((_unit,)))
        return CumulativeFunction(_cumul._args_, nargs=len(_cumul._args_))
    else:
        return CumulativeFunction(_cumul.args + [NegatedStepFunction((_unit,))])


def _subtract_unit_and_cumul(_unit, _cumul):
    # Nothing to be done: we need to make a new one because we can't prepend to
    # the list of args.
    return CumulativeFunction(
        [_unit] + [NegatedStepFunction((a,)) for a in _cumul.args]
    )


def _subtract_cumuls(_self, _other):
    if _self.nargs() == len(_self._args_):
        _self._args_.extend([NegatedStepFunction((a,)) for a in _other.args])
        return CumulativeFunction(_self._args_, nargs=len(_self._args_))
    else:
        # we have to clone the list of _args_
        return CumulativeFunction(
            _self.args + [NegatedStepFunction((a,)) for a in _other.args]
        )


def _generate_sum_expression(_self, _other):
    if isinstance(_self, CumulativeFunction):
        if isinstance(_other, CumulativeFunction):
            return _sum_cumuls(_self, _other)
        elif isinstance(_other, StepFunction):
            return _sum_cumul_and_unit(_self, _other)
    elif isinstance(_self, StepFunction):
        if isinstance(_other, CumulativeFunction):
            return _sum_unit_and_cumul(_self, _other)
        elif isinstance(_other, StepFunction):
            return _sum_two_units(_self, _other)
    raise TypeError(
        "Cannot add object of class %s to object of "
        "class %s" % (_other.__class__, _self.__class__)
    )


def _generate_difference_expression(_self, _other):
    if isinstance(_self, CumulativeFunction):
        if isinstance(_other, CumulativeFunction):
            return _subtract_cumuls(_self, _other)
        elif isinstance(_other, StepFunction):
            return _subtract_cumul_and_unit(_self, _other)
    elif isinstance(_self, StepFunction):
        if isinstance(_other, CumulativeFunction):
            return _subtract_unit_and_cumul(_self, _other)
        elif isinstance(_other, StepFunction):
            return _subtract_two_units(_self, _other)
    raise TypeError(
        "Cannot subtract object of class %s from object of "
        "class %s" % (_other.__class__, _self.__class__)
    )


class StepFunction(ExpressionBase):
    """
    The base class for the step function expression system.
    """

    __slots__ = ()
    PRECEDENCE = None

    def __add__(self, other):
        return _generate_sum_expression(self, other)

    def __radd__(self, other):
        # Mathematically this doesn't make a whole lot of sense, but we'll call
        # 0 a function and be happy so that sum() works as expected.
        if other == 0:
            return self
        return _generate_sum_expression(other, self)

    def __iadd__(self, other):
        return _generate_sum_expression(self, other)

    def __sub__(self, other):
        return _generate_difference_expression(self, other)

    def __rsub__(self, other):
        return _generate_difference_expression(other, self)

    def __isub__(self, other):
        return _generate_difference_expression(self, other)

    def within(self, bounds, times):
        return AlwaysIn(cumul_func=self, bounds=bounds, times=times)

    @property
    def args(self):
        return self._args_[: self.nargs()]


class Pulse(StepFunction):
    """
    A step function specified by an IntervalVar and an integer height that
    has value 0 before the IntervalVar's start_time and after the
    IntervalVar's end time and that takes the value specified by the 'height'
    during the IntervalVar. (These are often used to model resource
    constraints.)

    Args:
        interval_var (IntervalVar): the interval variable over which the
            pulse function is non-zero
        height (int): The value of the pulse function during the time
            interval_var is scheduled
    """

    __slots__ = '_args_'

    def __init__(self, args=None, interval_var=None, height=None):
        if args:
            if any(arg is not None for arg in (interval_var, height)):
                raise ValueError(
                    "Cannot specify both args and any of {interval_var, height}"
                )
            # Make sure this is a list because we may add to it if this is
            # summed with other StepFunctions
            self._args_ = [arg for arg in args]
        else:
            self._args_ = [interval_var, height]

        interval_var = self._args_[0]
        if (
            not isinstance(interval_var, IntervalVarData)
            or interval_var.ctype is not IntervalVar
        ):
            raise TypeError(
                "The 'interval_var' argument for a 'Pulse' must "
                "be an 'IntervalVar'.\n"
                "Received: %s" % type(interval_var)
            )

    @property
    def _interval_var(self):
        return self._args_[0]

    @property
    def _height(self):
        return self._args_[1]

    def nargs(self):
        return 2

    def _to_string(self, values, verbose, smap):
        return "Pulse(%s, height=%s)" % (values[0], values[1])


class Step(StepFunction):
    """
    A step function specified by a time point and an integer height that
    has value 0 before the time point and takes the value specified by the
    'height' after the time point.

    Args:
        time (IntervalVarTimePoint or int): the time point at which the step
            function becomes non-zero
        height (int): The value of the step function after the time point
    """

    __slots__ = '_args_'

    def __new__(cls, time, height):
        if isinstance(time, int):
            return StepAt((time, height))
        elif time.ctype is IntervalVarStartTime:
            return StepAtStart((time.get_associated_interval_var(), height))
        elif time.ctype is IntervalVarEndTime:
            return StepAtEnd((time.get_associated_interval_var(), height))
        else:
            raise TypeError(
                "The 'time' argument for a 'Step' must be either "
                "an 'IntervalVarTimePoint' (for example, the "
                "'start_time' or 'end_time' of an IntervalVar) or "
                "an integer time point in the time horizon.\n"
                "Received: %s" % type(time)
            )


class StepBase(StepFunction):
    __slots__ = '_args_'

    def __init__(self, args):
        # Make sure this is a list because we may add to it if this is summed
        # with otther StepFunctions
        self._args_ = [arg for arg in args]

    @property
    def _time(self):
        return self._args_[0]

    @property
    def _height(self):
        return self._args_[1]

    def nargs(self):
        return 2

    def _to_string(self, values, verbose, smap):
        return "Step(%s, height=%s)" % (values[0], values[1])


class StepAt(StepBase):
    __slots__ = ()


class StepAtStart(StepBase):
    __slots__ = ()

    def _to_string(self, values, verbose, smap):
        return "Step(%s, height=%s)" % (self._time.start_time, values[1])


class StepAtEnd(StepBase):
    __slots__ = ()

    def _to_string(self, values, verbose, smap):
        return "Step(%s, height=%s)" % (self._time.end_time, values[1])


class CumulativeFunction(StepFunction):
    """
    A sum of elementary step functions (Pulse and Step), defining a step
    function over time. (Often used to model resource constraints.)

    Args:
        args (list or tuple): Child elementary step functions of this node
    """

    __slots__ = ('_args_', '_nargs')
    PRECEDENCE = SumExpression.PRECEDENCE

    def __init__(self, args, nargs=None):
        # We make sure args are a list because we might add to them later, if
        # this is summed with another cumulative function.
        self._args_ = [arg for arg in args]
        if nargs is None:
            self._nargs = len(args)
        else:
            self._nargs = nargs

    def nargs(self):
        return self._nargs

    def _to_string(self, values, verbose, smap):
        s = ""
        for i, arg in enumerate(self.args):
            if isinstance(arg, NegatedStepFunction):
                s += str(arg) + ' '
            else:
                s += "+ %s "[2 * (i == 0) :] % str(arg)
        return s[:-1]


class NegatedStepFunction(StepFunction):
    """
    The negated form of an elementary step function: That is, it represents
    subtracting the elementary function's (nonnegative) height rather than
    adding it.

    Args:
       arg (Step or Pulse): Child elementary step function of this node
    """

    __slots__ = '_args_'

    def __init__(self, args):
        self._args_ = args

    def nargs(self):
        return 1

    def _to_string(self, values, verbose, smap):
        return "- %s" % values[0]


class AlwaysIn(BooleanExpression):
    """
    An expression representing the constraint that a cumulative function is
    required to take values within a tuple of bounds over a specified time
    interval. (Often used to enforce limits on resource availability.)

    Args:
        cumul_func (CumulativeFunction): Step function being constrained
        bounds (tuple of two integers): Lower and upper bounds to enforce on
            the cumulative function
        times (tuple of two integers): The time interval (start, end) over
            which to enforce the bounds on the values of the cumulative
            function.
    """

    __slots__ = ()

    def __init__(self, args=None, cumul_func=None, bounds=None, times=None):
        if args:
            if any(arg is not None for arg in {cumul_func, bounds, times}):
                raise ValueError(
                    "Cannot specify both args and any of {cumul_func, bounds, times}"
                )
            self._args_ = args
        else:
            self._args_ = (cumul_func, bounds[0], bounds[1], times[0], times[1])

    def nargs(self):
        return 5

    def _to_string(self, values, verbose, smap):
        return "%s.within(bounds=(%s, %s), times=(%s, %s))" % (
            values[0],
            values[1],
            values[2],
            values[3],
            values[4],
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.deprecation import relocated_module_attribute

relocated_module_attribute(
    'amplfunc_string_merge',
    'pyomo.solvers.amplfunc_merge.unique_paths',
    version='6.9.2',
    f_globals=globals(),
)


def unique_paths(*paths):
    """Merge paths, eliminating duplicates.

    Parameters
    ----------
    *path : str

        Each argument is a string containing one or more paths.
        Multiple libraries are separated by newlines.

    Returns
    -------
    str : merged list of newline-separated paths.

    """
    funcs = {}
    for src in paths:
        for line in src.splitlines():
            if not line:
                continue
            funcs[line] = None
    return "\n".join(funcs)


def amplfunc_merge(env, *funcs):
    """Merge AMPLFUNC and PYOMO_AMPLFUNC environment variables with provided values

    Paths are returned with entries from AMPLFUNC first, PYOMO_AMPLFUNC
    second, and the user-provided arguments last.  Duplicates and empty
    paths are filtered out.

    Parameters
    ----------
    env : Dict[str, str]
        Environment dictionary mapping environment variable names to values.

    *funcs : str
        Additional paths to combine.

    """
    return unique_paths(env.get("AMPLFUNC", ""), env.get("PYOMO_AMPLFUNC", ""), *funcs)
